#include "StdInc.h"
#include "RollEm.h"

#include "Combos.h"
#include "RollTypes.h"
#include "Strings.h"

Variable g_vPreCommonRoll [] = 
{
	"Roll Type",	DataTypeRollType,	COMMON_ROLLTYPE,	0,			0,			0,
	"Title",		DataTypeString,		0,					0,			1,			15,
	"",				DataTypeEnd,		0,					0,			0,			0,
};

Variable g_vNormalRoll [] = 
{
	"Number",		DataTypeNumber,		NORMAL_NUMBER,		1,			1,			1000,
	"Die Size",		DataTypeNumber,		NORMAL_DIE,			2,			1,			1000,
	"Modifier",		DataTypeNumber,		NORMAL_MODIFIER,	0,			-10000,		10000,
	"Best",			DataTypeNumber,		NORMAL_BEST,		0,			0,			1000,
	"",				DataTypeEnd,		0,					0,			0,			0,
};

Variable g_vSuccessRoll [] = 
{
	"Number",		DataTypeNumber,		SUCCESS_NUMBER,		1,			1,			1000,
	"Die Size",		DataTypeNumber,		SUCCESS_DIE,		2,			1,			1000,
	"Best",			DataTypeNumber,		SUCCESS_BEST,		0,			0,			1000,
	"Count",		DataTypeCompare,	SUCCESS_COMPARE,	0,			0,			0,
	"",				DataTypeNumber,		SUCCESS_VALUE,		1,			0,			1000,
	"Modifier",		DataTypeNumber,		SUCCESS_MODIFIER,	0,			-10000,		10000,
	"",				DataTypeEnd,		0,					0,			0,			0,
};

Variable g_vThacoRoll [] = 
{
	"Number",		DataTypeNumber,		THACO_NUMBER,		1,			1,			1000,
	"Die Size",		DataTypeNumber,		THACO_DIE,			20,			1,			1000,
	"Modifier",		DataTypeNumber,		THACO_MODIFIER,		0,			-10000,		10000,
	"Thac0",		DataTypeNumber,		THACO_VALUE,		20,			-1000,		1000,
	"",				DataTypeEnd,		0,					0,			0,			0,
};

Variable g_vOpenEndRoll [] = 
{
	"Die Size",		DataTypeNumber,		OPENEND_DIE,		100,		1,			1000,
	"Explode",		DataTypeCompare,	OPENEND_HIGHCOMPARE,0,			0,			0,
	"",				DataTypeNumber,		OPENEND_HIGHVALUE,	100,		1,			1000,
	"Implode",		DataTypeCompare,	OPENEND_LOWCOMPARE,	0,			0,			0,
	"",				DataTypeNumber,		OPENEND_LOWVALUE,	1,			0,			1000,
	"Modifier",		DataTypeNumber,		OPENEND_MODIFIER,	0,			-10000,		10000,
	"Max Rolls",	DataTypeNumber,		OPENEND_MAXROLLS,	100,		2,			1000,
	"",				DataTypeEnd,		0,					0,			0,			0,
};


Variable g_vAdditiveRoll [] = 
{
	"1st Number",	DataTypeNumber,		ADDITIVE_NUMBER1,	1,			1,			1000,
	"1st Die Size",	DataTypeNumber,		ADDITIVE_DIE1,		2,			1,			1000,
	"2nd Number",	DataTypeNumber,		ADDITIVE_NUMBER2,	1,			1,			1000,
	"2nd Die Size",	DataTypeNumber,		ADDITIVE_DIE2,		2,			1,			1000,
	"Modifier",		DataTypeNumber,		ADDITIVE_MODIFIER,	0,			-10000,		10000,
	"",				DataTypeEnd,		0,					0,			0,			0,
};

Variable g_vPostCommonRoll [] = 
{
	"Beep",			DataTypeSound,		COMMON_BEEPTYPE1,	0,			0,			0,
	"",				DataTypeCompare,	COMMON_BEEPCOMPARE1,0,			0,			0,
	"",				DataTypeNumber,		COMMON_BEEPVALUE1,	0,			-1000000,	1000000,
	"Beep",			DataTypeSound,		COMMON_BEEPTYPE2,	0,			0,			0,
	"",				DataTypeCompare,	COMMON_BEEPCOMPARE2,0,			0,			0,
	"",				DataTypeNumber,		COMMON_BEEPVALUE2,	0,			-1000000,	1000000,
	"",				DataTypeEnd,		0,					0,			0,			0,
};

VariableNode g_vRolls [] = 
{
	"Normal", g_vNormalRoll,
	"Success Count", g_vSuccessRoll,
	"Open Ended", g_vOpenEndRoll,
	"Thac0", g_vThacoRoll,
	"Additive", g_vAdditiveRoll,
	"Subtractive", g_vAdditiveRoll,
};

DieRoll g_vAdvDieRoll;
Int16 g_iAdvPos;
Int16 g_iSoundCount, g_iCompareCount;
Int16 g_iPreCommonCount, g_iRollTypeCount, g_iPostCommonCount;
Variable *g_pRollType;

Boolean AdvCustomProc (EventPtr pEvent);

void ConstructorForm (FormType *pForm);
void DestructorForm (FormType *pForm);
void DrawForm (FormType *pForm);
void LoadAdvCustomTable (FormType *pForm);
void TapAdvCustomTable (FormType *pForm, TablePtr pTable, UInt16 wRow, UInt16 wColumn);
void AdvCustomDrawFunc (void *pTable, Int16 wRow, Int16 wColumn, RectangleType *pRect);
void LoadDefaults (FormType *pForm);
Variable* GetVariableAtPos(Int16 iPos);

void RollTypeString (Int16 iIndex, Char *pString)
{
	StrCopy(pString, g_vRolls [iIndex].pFieldName);
}

void CompareString (Int16 iIndex, Char *pString)
{
	Char *szCompare [] = { "=", ">", "<", ">=", "<=", "!=" };

	StrCopy(pString, szCompare[iIndex]);
}

void SoundString (Int16 iIndex, Char *pString)
{
	MemHandle hRecord;
	MemPtr pRecord;
	Char *pChar;
	Int16 i, iCount;

	// The first string is none.
	if (iIndex == 0)
	{
		StrCopy(pString, "None");
		return;
	}

	iCount = 0;
	g_vDataRecordInfo.vId = RECORDID_SOUNDS;
	g_vDataRecordInfo.vSubId = RECORDID_NONE;
	if (ReadRecordInfo(&g_vDataRecordInfo))
	{
		pChar = (Char*)g_vDataRecordInfo.vDatabase.pRecord;
		iCount = *((UInt16*)pChar) + 1;

		// Moveit 
		pChar += sizeof(UInt16);
		for (i=1; i<iCount; i++)
		{
			// Skip the sounds.
			while (((NotePtr)pChar)->uiDuration)
				pChar += sizeof(Note);

			// Move up.
			pChar += sizeof(Note);

			if (iIndex == i)
			{
				// Make and copy the title.
				StrCopy(pString, pChar);
				break;
			}

			// Skip the titles.
			pChar += MAX_TITLE_LEN + 1;
		}
		CloseRecordInfo(&g_vDataRecordInfo);
	}
}

void DoModalAdvCustom (DieRollPtr pDieRoll)
{
	FormType *pForm, *pOldForm;
	UInt16 wResult;

	wResult = 0;
	pOldForm = FrmGetActiveForm();
	if (pForm = FrmInitForm(IDF_ADVCUSTOM))
	{
		// Pre form stuff.
		FrmSetActiveForm(pForm);
		FrmDrawForm(pForm);

		ConstructorForm(pForm);
		MemMove(&g_vAdvDieRoll, pDieRoll, sizeof(DieRoll));
		DrawForm(pForm);

		FrmSetEventHandler(pForm, (FormEventHandlerPtr) AdvCustomProc);
		wResult = FrmDoDialog(pForm);

		if (wResult == IDC_OK)
		{
			// Copy the new information.
			MemMove(pDieRoll, &g_vAdvDieRoll, sizeof(DieRoll));

			// Make a title if we don't have one.
			if (StrLen(pDieRoll->szTitle) < 1)
			{
				CreateTitle(pDieRoll, pDieRoll->szTitle, MAX_TITLE_LEN+1);
			}
		}

		// We now return to our other window.
		if (pOldForm)
			FrmSetActiveForm(pOldForm);
		FrmDeleteForm(pForm);
	}
}

Boolean AdvCustomProc (EventPtr pEvent)
{
	FormType *pForm = FrmGetActiveForm();
	Boolean bHandled = false;

	switch (pEvent->eType) 
	{
	case frmCloseEvent: // The form is closing.
	{
		DestructorForm(pForm);
		bHandled = false;
		break;
	}

	case frmUpdateEvent:
	{
		FrmDrawForm(pForm);
		DrawForm(pForm);
		bHandled = true;
		break;
	}

	case ctlSelectEvent:  // A control button was pressed and released.
	{
		switch (pEvent->data.ctlEnter.controlID)
		{
		case IDC_OK:
		case IDC_CANCEL:
			bHandled = false;
			break;
		default:;
		}
		break;
	}

	case sclRepeatEvent:
	{
		bHandled = false;
		switch (pEvent->data.sclRepeat.scrollBarID)
		{
		case IDC_ADV_CUSTOM_SCROLL:
			g_iAdvPos = pEvent->data.sclRepeat.newValue;
			LoadAdvCustomTable(pForm);
			break;
		default:;
		}
		break;
	}

	case tblSelectEvent:  // A control button was pressed and released.
	{
		TapAdvCustomTable(pForm, pEvent->data.tblSelect.pTable, pEvent->data.tblSelect.row, pEvent->data.tblSelect.column);
		bHandled = true;
		break;
	}

    case menuEvent:
	{
        switch (pEvent->data.menu.itemID)
        {
		case IDM_FORM_DEFAULT:
			LoadDefaults(pForm);
			bHandled = true;
			break;
		}
		break;
	}

	default:
		bHandled = false;
	}

    return bHandled;
}

void ConstructorForm (FormType *pForm)
{
	g_iAdvPos = 0;
	g_iSoundCount = 1;
	g_iCompareCount = 6;

	g_vDataRecordInfo.vId = RECORDID_SOUNDS;
	g_vDataRecordInfo.vSubId = RECORDID_NONE;
	if (ReadRecordInfo(&g_vDataRecordInfo))
	{
		g_iSoundCount += *((UInt16*)g_vDataRecordInfo.vDatabase.pRecord);
		CloseRecordInfo(&g_vDataRecordInfo);
	}
}

void DestructorForm (FormType *pForm)
{
}

void DrawForm (FormType *pForm)
{
	LoadAdvCustomTable(pForm);
}

void LoadAdvCustomTable (FormType *pForm)
{
	Int16 iRows, iIndex, iCount;
	ScrollBarPtr pScroll;
	TablePtr pTable;
	Variable *pVariable;
	
	pScroll = FrmGetObjectPtr(pForm, FrmGetObjectIndex(pForm, IDC_ADV_CUSTOM_SCROLL));
	pTable = FrmGetObjectPtr(pForm, FrmGetObjectIndex(pForm, IDC_ADV_CUSTOM_LIST));

	iRows = TblGetNumberOfRows(pTable);
	TblUnhighlightSelection(pTable);

	g_pRollType = g_vRolls [g_vAdvDieRoll.lField[COMMON_ROLLTYPE]].pVariable;

	g_iPreCommonCount = 0;
	g_iRollTypeCount = 0;
	g_iPostCommonCount = 0;
	iIndex = 0;

	// Also count the common stuff
	pVariable = g_vPreCommonRoll;
	while (pVariable->vDataType != DataTypeEnd)
	{
		g_iPreCommonCount++;
		pVariable++;
	}

	// Count of the items.
	pVariable = g_pRollType;
	while (pVariable->vDataType != DataTypeEnd)
	{
		g_iRollTypeCount++;
		pVariable++;
	}

	// Also count the common stuff
	pVariable = g_vPostCommonRoll;
	while (pVariable->vDataType != DataTypeEnd)
	{
		g_iPostCommonCount++;
		pVariable++;
	}

	// Add all of them to make the total.
	iCount = g_iPreCommonCount + g_iRollTypeCount + g_iPostCommonCount;

	// Insert the items into the table.
	for (iIndex=0; (iIndex<iRows) && (iIndex<iCount); iIndex++)
	{
		TblSetItemStyle(pTable, iIndex, 0, customTableItem);
		TblSetItemInt(pTable, iIndex, 0, iIndex + g_iAdvPos);
		TblMarkRowInvalid(pTable, iIndex);
		TblSetRowUsable(pTable, iIndex, true);
	}

	// Disable the unused rows.
	for (; iIndex<iRows; iIndex++)
	{
		TblSetRowUsable(pTable, iIndex, false);
		TblMarkRowInvalid(pTable, iIndex);
	}
	
	TblSetColumnUsable(pTable, 0, true);
	TblSetCustomDrawProcedure(pTable, 0, AdvCustomDrawFunc);
	TblRedrawTable(pTable);

	SclSetScrollBar (pScroll, g_iAdvPos, 0, ((iCount > iRows) ? (iCount - iRows) : 0), iRows);
}

void AdvCustomDrawFunc (void *pTable, Int16 wRow, Int16 wColumn, RectangleType *pRect)
{
	Char szBuffer [32];
	Variable *pVariable;
	Int16 iPos, iCount;

	iPos = TblGetItemInt(pTable, wRow, wColumn);
	pVariable = GetVariableAtPos(iPos);

	switch(pVariable->vDataType)
	{
	case DataTypeNumber:
		StrLToAComma(szBuffer, g_vAdvDieRoll.lField[pVariable->vDataPos]);
		break;
	case DataTypeString:
		StrCopy(szBuffer, g_vAdvDieRoll.szTitle);
		break;
//	case DataTypeCheck:
//		break;
	case DataTypeRollType:
		if (g_vAdvDieRoll.lField[COMMON_ROLLTYPE] >= RollTypeEnd)
			g_vAdvDieRoll.lField[COMMON_ROLLTYPE] = RollTypeNormal;
		RollTypeString(g_vAdvDieRoll.lField[COMMON_ROLLTYPE], szBuffer);
		break;
	case DataTypeCompare:
		if (g_vAdvDieRoll.lField[pVariable->vDataPos] >= g_iCompareCount)
			g_vAdvDieRoll.lField[pVariable->vDataPos] = 0;
		CompareString(g_vAdvDieRoll.lField[pVariable->vDataPos], szBuffer);
		break;
	case DataTypeSound:
		if (g_vAdvDieRoll.lField[pVariable->vDataPos] >= g_iSoundCount)
			g_vAdvDieRoll.lField[pVariable->vDataPos] = 0;
		SoundString(g_vAdvDieRoll.lField[pVariable->vDataPos], szBuffer);
		break;
	case DataTypeRollLabel:
		DrawCharacters(pVariable->szFieldName, StrLen(pVariable->szFieldName), pRect, DC_CENTER, true);
		return;
	default:
		*szBuffer = '\0';
	}

	InflateRect(pRect, -3, 0);
	DrawCharacters(pVariable->szFieldName, StrLen(pVariable->szFieldName), pRect, DC_LEFT, true);
	DrawCharacters(szBuffer, StrLen(szBuffer), pRect, DC_RIGHT, true);
}

void TapAdvCustomTable (FormType *pForm, TablePtr pTable, UInt16 wRow, UInt16 wColumn)
{
	Variable *pVariable, *pTitleVariable;
	Char *pTitle;
	Int16 iPos;

	iPos = TblGetItemInt(pTable, wRow, wColumn);
	pVariable = GetVariableAtPos(iPos);
	pTitleVariable = pVariable;

	// Find a title.
	pTitle = pTitleVariable->szFieldName;
	while (*pTitle == '\0')
	{
		pTitleVariable--;
		pTitle = pTitleVariable->szFieldName;
	}

	switch(pVariable->vDataType)
	{
	case DataTypeNumber:
		DoModalNumbers(pTitle, &g_vAdvDieRoll.lField[pVariable->vDataPos], pVariable->lMin, pVariable->lMax);
		break;
	case DataTypeString:
		DoStringsModal(pTitle, g_vAdvDieRoll.szTitle, MAX_TITLE_LEN);
		break;
//	case DataTypeCheck:
//		break;
	case DataTypeRollType:
		if (DoCombosModal(pTitle, &g_vAdvDieRoll.lField[COMMON_ROLLTYPE], 0, RollTypeEnd, RollTypeString) == IDC_OK)
		{
			LoadDefaults(pForm);
		}
		break;
	case DataTypeCompare:
		DoCombosModal(pTitle, &g_vAdvDieRoll.lField[pVariable->vDataPos], 0, g_iCompareCount, CompareString);
		break;
	case DataTypeSound:
		DoCombosModal(pTitle, &g_vAdvDieRoll.lField[pVariable->vDataPos], 0, g_iSoundCount, SoundString);
		break;
	case DataTypeRollLabel:
	default:;
	}
	TblUnhighlightSelection(pTable);
	TblMarkRowInvalid(pTable, wRow);
	TblRedrawTable(pTable);
}

void LoadDefaults (FormType *pForm)
{
	Variable *pVariable;
	
	// Set the defaults for the non-common stuff.
	pVariable = g_vRolls [g_vAdvDieRoll.lField[COMMON_ROLLTYPE]].pVariable;
	while (pVariable->vDataType != DataTypeEnd)
	{
		g_vAdvDieRoll.lField[pVariable->vDataPos] = pVariable->lDefault;
		pVariable++;
	}

	LoadAdvCustomTable(pForm);
}

Variable* GetVariableAtPos(Int16 iPos)
{
	if (iPos < g_iPreCommonCount)
		return &g_vPreCommonRoll [iPos];

	iPos -= g_iPreCommonCount;
	if (iPos < g_iRollTypeCount)
		return &g_pRollType [iPos];

	return &g_vPostCommonRoll [iPos - g_iRollTypeCount];
}
