#include "StdInc.h"
#include "RollEm.h"

#include "Display.h"
#include "DisplayItem.h"

// ************************************************************* //

#define SELECT_NONE		-2
#define SELECT_EMPTY	-1

// ************************************************************* //

StackNode g_vDisplayItems;
RectangleType g_vLastSelectedItem;
Int16 g_iLastSelectedItem;
Char g_szLabel [MAX_DISPLAY_TITLE];

Char *g_szTypes [] = 
{
	"Results",
    "Clear",
    "Button 1",
    "Button 2",			
    "Button 3",			
    "Button 4",			
    "Button 5",			
    "Button 6",			
    "Button 7",			
    "Button 8",			
	"Prev Page",		
    "Next Page",		
    "Page Number",		
    "Custom",	
    "Type",		
    "Total",		
    "Sum",			
    "Mean",		
    "Median",		
    "Mode",		
    "Sub Rolls",	
    "Sub Dice",		
    "Sub Total",	
};

// ************************************************************* //

void InitializeDisplay ();
void DeinitializeDisplay ();

void ReadDisplays ();
void WriteDisplays ();
void ClearDisplays ();

void RedrawDisplay (FormPtr pForm);

Boolean DisplayTap (FormPtr pForm, EventPtr pEvent);
void SetSelection (RectangleType *pRect, Int16 iItem);
void DrawSelection ();

// ************************************************************* //

Boolean DisplayProc(EventPtr pEvent)
{
	FormPtr pForm = FrmGetActiveForm();
	Boolean bHandled = false;

    switch (pEvent->eType) 
    {
    case frmOpenEvent:
    {
		InitializeDisplay();
		bHandled = true;
        break;
    }

    case frmCloseEvent:
    {
		DeinitializeDisplay();
		bHandled = false;
        break;
    }

	case frmUpdateEvent:
	{
		FrmDrawForm(pForm);
		RedrawDisplay(pForm);
		bHandled = true;
		break;
	}

	case ctlSelectEvent:  // A control button was pressed and released.
	{
        switch (pEvent->data.ctlEnter.controlID)
        {
        case IDC_OK:
			WriteDisplays();
        case IDC_CANCEL:
            FrmReturnToForm(0);
			DeinitializeDisplay();
			RefreshForm();
            break;
        }
        bHandled = true;
        break;
	}

    case menuEvent:
	{
        switch (pEvent->data.menu.itemID)
        {
		case IDM_ITEM_ADD:
		{
			if (g_iLastSelectedItem != SELECT_EMPTY)
			{
				FrmCustomAlert(IDA_NEED_SELECTION, "an empty area", "add item", NULL);
			}
			else
			{
				DisplayNode *pDisplayNode;
				Int16 iDisplayCount;
				Char szLabel [MAX_DISPLAY_TITLE];
				Int16 iType = 0;

				*szLabel = '\0';
				if (DoDisplayItemModal(szLabel, &iType, g_szTypes, 23) == IDC_OK)
				{
					iDisplayCount = StackGetCount(&g_vDisplayItems);
					StackAddRear(&g_vDisplayItems, 0, sizeof(DisplayNode));
					pDisplayNode = (DisplayNode*)StackGetData(&g_vDisplayItems, iDisplayCount);

					MemMove(&pDisplayNode->vCtrlRect, &g_vLastSelectedItem, sizeof(RectangleType));
					MemSet(pDisplayNode->szTitle, MAX_DISPLAY_TITLE, '\0');
					StrCopy(pDisplayNode->szTitle, szLabel);
					pDisplayNode->vType = iType;

					SetSelection(0, SELECT_NONE);
					RefreshForm();
				}
			}
			break;
		}

		case IDM_ITEM_DELETE:
		{
			if ((g_iLastSelectedItem == SELECT_NONE) || (g_iLastSelectedItem == SELECT_EMPTY))
			{
				FrmCustomAlert(IDA_NEED_SELECTION, "an item", "delete", NULL);
			}
			else
			{
				if (!FrmCustomAlert(IDA_DELETE_CONFIRMATION, "item", NULL, NULL))
				{
					StackDelete(&g_vDisplayItems, g_iLastSelectedItem);
					SetSelection(0, SELECT_NONE);
				}
				RefreshForm();
			}
			break;
		}

		case IDM_ITEM_EDIT:
		{
			if ((g_iLastSelectedItem == SELECT_NONE) || (g_iLastSelectedItem == SELECT_EMPTY))
			{
				FrmCustomAlert(IDA_NEED_SELECTION, "an item", "edit", NULL);
			}
			else
			{
				DisplayNode *pDisplayNode;

				pDisplayNode = StackGetData(&g_vDisplayItems, g_iLastSelectedItem);
				DoDisplayItemModal(pDisplayNode->szTitle, &pDisplayNode->vType, g_szTypes, 23);
				RefreshForm();
			}
			break;
		}

		case IDM_FORM_CLEAR:
		{
			if (!FrmCustomAlert(IDA_FORM_CLEAR, "items", NULL, NULL))
			{
				ClearDisplays();
				RefreshForm();
			}
			break;
		}

		case IDM_FORM_DEFAULT:
		{
			if (!FrmCustomAlert(IDA_FORM_DEFAULT, "items", NULL, NULL))
			{
				ClearDisplays();
				LoadDefaultDisplays();
				ReadDisplays();
				RefreshForm();
			}
			break;
		}

		default:;
        }
        bHandled = true;
        break;
	}

	case penDownEvent:
	{
		bHandled = DisplayTap(pForm, pEvent);
		break;
	}

	default:
		bHandled = false;
    }

    return bHandled;
}

void LoadDefaultDisplays ()
{
	MemHandle hDisplayItem;

	// Get the data from the resource.
	hDisplayItem = DmGet1Resource('Data', IDR_DISPLAYS);
	if (hDisplayItem != 0)
	{
		g_vDataRecordInfo.vId = RECORDID_DISPLAYS;
		g_vDataRecordInfo.vSubId = 0;
		g_vDataRecordInfo.vInput.vOffset = 0;
		g_vDataRecordInfo.vInput.pRecord = MemHandleLock(hDisplayItem);
		g_vDataRecordInfo.vInput.vLength = MemHandleSize(hDisplayItem);

		if (OverwriteRecordInfo(&g_vDataRecordInfo))
		{
			CloseRecordInfo(&g_vDataRecordInfo);
		}

		// Release the data.
		MemHandleUnlock(hDisplayItem);
		DmReleaseResource(hDisplayItem);
	}
}

void InitializeDisplay ()
{
	g_iLastSelectedItem = SELECT_NONE;
	MemSet(&g_vLastSelectedItem, sizeof(RectangleType), '\0');

	// Do the every time initialization 
	//  of the Sound dialog here.
	StackConstruct(&g_vDisplayItems);
	ReadDisplays();
	RefreshForm();
}

void DeinitializeDisplay ()
{
	ClearDisplays();
}

void ReadDisplays ()
{
	Int16 iDisplayCount, i;
	DisplayNode *pDisplayNode, *pRecordNode;

	g_vDataRecordInfo.vId = RECORDID_DISPLAYS;
	g_vDataRecordInfo.vSubId = 0;
	if (ReadRecordInfo(&g_vDataRecordInfo))
	{
		iDisplayCount = g_vDataRecordInfo.vDatabase.vLength / sizeof(DisplayNode);
		pRecordNode = (DisplayNode*)((Char*)g_vDataRecordInfo.vDatabase.pRecord + MAX_DISPLAY_TITLE);
		StrCopy(g_szLabel, (Char*)g_vDataRecordInfo.vDatabase.pRecord);
		for (i=0; i<iDisplayCount; i++)
		{
			// Add a new item to the list.
			StackAddRear(&g_vDisplayItems, 0, sizeof(DisplayNode));
			pDisplayNode = (DisplayNode*)StackGetData(&g_vDisplayItems, i);

			MemMove(pDisplayNode, pRecordNode, sizeof(DisplayNode));

			// Convert the data
			pDisplayNode->vCtrlRect.topLeft.x = ((pDisplayNode->vCtrlRect.topLeft.x / DISPLAY_WIDTH) * PREVIEW_WIDTH) + PREVIEW_OFFSET_X;
			pDisplayNode->vCtrlRect.topLeft.y = ((pDisplayNode->vCtrlRect.topLeft.y / DISPLAY_HEIGHT) * PREVIEW_HEIGHT) + PREVIEW_OFFSET_Y;
			pDisplayNode->vCtrlRect.extent.x = ((pDisplayNode->vCtrlRect.extent.x / DISPLAY_WIDTH) * PREVIEW_WIDTH) + 1;
			pDisplayNode->vCtrlRect.extent.y = ((pDisplayNode->vCtrlRect.extent.y / DISPLAY_HEIGHT) * PREVIEW_HEIGHT) + 1;
			InflateRect(&pDisplayNode->vCtrlRect, -1, -1);

			pRecordNode++;
		}
		CloseRecordInfo(&g_vDataRecordInfo);
	}
}

void WriteDisplays ()
{
	Int16 iDisplayCount, i;
	DisplayNode *pDisplayNode;

	iDisplayCount = StackGetCount(&g_vDisplayItems);

	g_vDataRecordInfo.vId = RECORDID_DISPLAYS;
	g_vDataRecordInfo.vSubId = 0;
	g_vDataRecordInfo.vInput.pRecord = 0;
	g_vDataRecordInfo.vInput.vLength = (sizeof(DisplayNode) * iDisplayCount) + MAX_DISPLAY_TITLE;

	if (OverwriteRecordInfo(&g_vDataRecordInfo))
	{
		// Write the title.
		g_vDataRecordInfo.vInput.vOffset = 0;
		g_vDataRecordInfo.vInput.pRecord = g_szLabel;
		g_vDataRecordInfo.vInput.vLength = MAX_DISPLAY_TITLE;
		WriteRecordInfo(&g_vDataRecordInfo);

		for (i=0; i<iDisplayCount; i++)
		{
			pDisplayNode = (DisplayNode*)StackGetData(&g_vDisplayItems, i);

			pDisplayNode->vCtrlRect.topLeft.x = (((pDisplayNode->vCtrlRect.topLeft.x - 1) - PREVIEW_OFFSET_X) / PREVIEW_WIDTH) * DISPLAY_WIDTH;
			pDisplayNode->vCtrlRect.topLeft.y = (((pDisplayNode->vCtrlRect.topLeft.y - 1) - PREVIEW_OFFSET_Y) / PREVIEW_HEIGHT) * DISPLAY_HEIGHT;
			pDisplayNode->vCtrlRect.extent.x = (((pDisplayNode->vCtrlRect.extent.x + 2) / PREVIEW_WIDTH) * DISPLAY_WIDTH) + 1;
			pDisplayNode->vCtrlRect.extent.y = (((pDisplayNode->vCtrlRect.extent.y + 2) / PREVIEW_HEIGHT) * DISPLAY_HEIGHT) + 1;

			// Write the item.
			g_vDataRecordInfo.vInput.vOffset += g_vDataRecordInfo.vInput.vLength;
			g_vDataRecordInfo.vInput.pRecord = pDisplayNode;
			g_vDataRecordInfo.vInput.vLength = sizeof(DisplayNode);
			WriteRecordInfo(&g_vDataRecordInfo);
		}
		CloseRecordInfo(&g_vDataRecordInfo);
	}
}

void ClearDisplays ()
{
	StackDestruct(&g_vDisplayItems);
}

void RedrawDisplay (FormPtr pForm)
{
	DrawDisplayGrid(pForm, PREVIEW_OFFSET_X, PREVIEW_OFFSET_Y, PREVIEW_WIDTH, PREVIEW_HEIGHT);
}

void DrawDisplayGrid (FormPtr pForm, Int16 iOffsetX, Int16 iOffsetY, Int16 iWidth, Int16 iHeight)
{
	Int16 x1, y1, x2, y2, i, iDisplayCount;
	RectangleType vCtrlRect;
	DisplayNode *pDisplayNode;
	FontID vFontId;

	FrmGetObjectBounds(pForm, FrmGetObjectIndex(pForm, IDC_RESULT), &vCtrlRect);
	WinEraseRectangle(&vCtrlRect, 0);

	x1 = iOffsetX;
	y1 = iOffsetY;
	x2 = x1 + (25 * iWidth);

	// Draw the horizontal lines
	for (i=0; i<24; i++)
	{
		WinDrawLine(x1, y1, x2, y1);
		y1 += iHeight;
	}

	y1 = iOffsetY;
	y2 = y1 + (23 * iHeight);

	// Draw the vertical lines.
	for (i=0; i<26; i++)
	{
		WinDrawLine(x1, y1, x1, y2);
		x1 += iWidth;
	}

	// Display the items.
	iDisplayCount = StackGetCount(&g_vDisplayItems);
	for (i=0; i<iDisplayCount; i++)
	{
		pDisplayNode = (DisplayNode*)StackGetData(&g_vDisplayItems, i);
		switch(pDisplayNode->vType)
		{
		case DisplayItemButtonRolls:
			DrawButtonValue(pForm, "R", &pDisplayNode->vCtrlRect, false);
			break;
		case DisplayItemButtonClear:
			DrawButtonValue(pForm, "C", &pDisplayNode->vCtrlRect, false);
			break;
		case DisplayItemButton0:
		case DisplayItemButton1:
		case DisplayItemButton2:
		case DisplayItemButton3:
		case DisplayItemButton4:
		case DisplayItemButton5:
		case DisplayItemButton6:
		case DisplayItemButton7:
			DrawButtonValue(pForm, "Button", &pDisplayNode->vCtrlRect, false);
			break;
		case DisplayItemButtonNext:
			vFontId = FntSetFont(3);
			DrawButtonValue(pForm, "\004", &pDisplayNode->vCtrlRect, false);
			FntSetFont(vFontId);
			break;
		case DisplayItemButtonPrev:
			vFontId = FntSetFont(3);
			DrawButtonValue(pForm, "\003", &pDisplayNode->vCtrlRect, false);
			FntSetFont(vFontId);
			break;
		case DisplayItemButtonPage:
			vFontId = FntSetFont(1);
			DrawStringValue(pForm, 0, "1", &pDisplayNode->vCtrlRect, false);
			FntSetFont(vFontId);
			break;
		case DisplayItemButtonCustom:
			vFontId = FntSetFont(1);
			DrawButtonValue(pForm, "Cancel", &pDisplayNode->vCtrlRect, false);
			FntSetFont(vFontId);
			break;
			break;
		case DisplayItemRollType:
			DrawStringValue(pForm, 0, "Type", &pDisplayNode->vCtrlRect, false);
			break;
		case DisplayItemRollTotal:
		case DisplayItemRollSum:
		case DisplayItemRollMean:
		case DisplayItemRollMedian:
		case DisplayItemRollMode:
		case DisplayItemRollSubDice:
		case DisplayItemRollSubRolls:
		case DisplayItemRollSubTotal:
			DrawStringValue(pForm, pDisplayNode->szTitle, "", &pDisplayNode->vCtrlRect, false);
			break;
		default:
			WinInvertRectangle(&pDisplayNode->vCtrlRect, 0x0000);
		}
	}
	DrawSelection();
}

void DrawStringValue (FormPtr pForm, Char *pLabel, Char *pValue, RectangleType *pCtrlRect, Boolean bSelected)
{
	Boolean bVersion350 = IsRomVersion(VERSION_350);

	if (bVersion350 == true)
	{
		WinSetTextColor(UIColorGetTableEntryIndex(((bSelected == true) ? UIObjectSelectedForeground : UIObjectForeground)));
		WinSetBackColor(UIColorGetTableEntryIndex(((bSelected == true) ? UIObjectSelectedFill : UIObjectFill)));
	}
	else if (bSelected == true)
	{
		WinInvertRectangle(pCtrlRect, 0x00);
		return;
	}

	WinEraseRectangle(pCtrlRect, 0x00);
	if (pLabel == 0)
	{
		DrawCharacters(pValue, StrLen(pValue), pCtrlRect, DC_CENTER, true); 
	}
	else
	{
		DrawCharacters(pLabel, StrLen(pLabel), pCtrlRect, DC_LEFT, true); 
		DrawCharacters(pValue, StrLen(pValue), pCtrlRect, DC_RIGHT, true); 
	}
}

void DrawButtonValue (FormPtr pForm, Char *pText, RectangleType *pCtrlRect, Boolean bSelected)
{
	Boolean bVersion350 = IsRomVersion(VERSION_350);

	if (bVersion350 == true)
	{
		WinSetTextColor(UIColorGetTableEntryIndex(((bSelected == true) ? UIObjectSelectedForeground : UIObjectForeground)));
		WinSetBackColor(UIColorGetTableEntryIndex(((bSelected == true) ? UIObjectSelectedFill : UIObjectFill)));
	}
	else if (bSelected == true)
	{
		InflateRect(pCtrlRect, -1, -1);
		WinInvertRectangle(pCtrlRect, 0x04);
		InflateRect(pCtrlRect, 1, 1);
		return;
	}

	InflateRect(pCtrlRect, -1, -1);
	WinDrawRectangleFrame(0x0401, pCtrlRect);
	WinEraseRectangle(pCtrlRect, 0x04);
	InflateRect(pCtrlRect, -2, 0);
	DrawCharacters(pText, StrLen(pText), pCtrlRect, DC_CENTER, true); 
	InflateRect(pCtrlRect, 3, 1);
}

Boolean DisplayTap (FormPtr pForm, EventPtr pEvent)
{
	RectangleType vCtrlRect, vRect;
	Int16 x, y, i;
	Boolean bPenDown;
	DisplayNode *pDisplayNode;
	Int16 iDisplayCount = 0;

	// Are we within the the gadget?
	FrmGetObjectBounds(pForm, FrmGetObjectIndex(pForm, IDC_RESULT), &vCtrlRect);
	if (!RctPtInRectangle(pEvent->screenX, pEvent->screenY, &vCtrlRect))
		return false;

	// Find the controls the wen already have.
	iDisplayCount = StackGetCount(&g_vDisplayItems);
	for (i=0; i<iDisplayCount; i++)
	{
		pDisplayNode = (DisplayNode*)StackGetData(&g_vDisplayItems, i);

		InflateRect(&pDisplayNode->vCtrlRect, 1, 1);
		if (RctPtInRectangle(pEvent->screenX, pEvent->screenY, &pDisplayNode->vCtrlRect))
		{
			InflateRect(&pDisplayNode->vCtrlRect, -1, -1);
			SetSelection(&pDisplayNode->vCtrlRect, i);
			return true;
		}
		InflateRect(&pDisplayNode->vCtrlRect, -1, -1);
	}

	// Clear the last selection.
	SetSelection(0, SELECT_NONE);

	vRect.topLeft.x = (((pEvent->screenX - PREVIEW_OFFSET_X) / PREVIEW_WIDTH) * PREVIEW_WIDTH) + PREVIEW_OFFSET_X + 1;
	vRect.topLeft.y = (((pEvent->screenY - PREVIEW_OFFSET_Y) / PREVIEW_HEIGHT) * PREVIEW_HEIGHT) + PREVIEW_OFFSET_Y + 1;
	vRect.extent.x = PREVIEW_WIDTH - 1;
	vRect.extent.y = PREVIEW_HEIGHT - 1;

	WinInvertRectangle(&vRect, 0);

	do 
	{
		PenGetPoint(&x, &y, &bPenDown);
		if ((vRect.topLeft.x < x) && (vRect.topLeft.y < y) && (RctPtInRectangle(x, y, &vCtrlRect) == true))
		{
			x = ((((x - vRect.topLeft.x) / PREVIEW_WIDTH) + 1) * PREVIEW_WIDTH) - 1;
			y = ((((y - vRect.topLeft.y) / PREVIEW_HEIGHT) + 1) * PREVIEW_HEIGHT) - 1;
			if ((x != vRect.extent.x) || (y != vRect.extent.y))
			{
				WinInvertRectangle(&vRect, 0);
				vRect.extent.x = x;
				vRect.extent.y = y;
				WinInvertRectangle(&vRect, 0);
			}
		}
	} while (bPenDown);

	WinInvertRectangle(&vRect, 0);

	// Clear the last selection.
	SetSelection(&vRect, SELECT_EMPTY);

	return true;
}

void SetSelection (RectangleType *pRect, Int16 iItem)
{
	DrawSelection();
	
	if (pRect == 0)
	{
		g_iLastSelectedItem = iItem;
		MemSet(&g_vLastSelectedItem, sizeof(RectangleType), '\0');
	}
	else
	{
		g_iLastSelectedItem = iItem;
		MemMove(&g_vLastSelectedItem, pRect, sizeof(RectangleType));
	}

	DrawSelection();
}

void DrawSelection ()
{
	if ((g_vLastSelectedItem.extent.x != 0) && (g_vLastSelectedItem.extent.y != 0))
	{
		WinInvertRectangle(&g_vLastSelectedItem, 0);
	}
}
