#include "StdInc.h"
#include "RollEm.h"

#include "AdvCustom.h"
#include "Display.h"
#include "Prefs.h"
#include "RollFunc.h"
#include "RollResults.h"
#include "RollTypes.h"
#include "Sound.h"

#define DRAW_ALL		0xffff
#define DRAW_PAGE		0x0001
#define DRAW_ROLL		0x0002

Int16 StartApplication();
Int16 StopApplication();
void EventLoop();

Boolean OpenDatabases();
Boolean CreateDatabases();
Boolean CloseDatabases();
Boolean InitDatabases();

Boolean MainProc(EventPtr event);

void ShowAbout ();

void NextPage ();
void PrevPage ();
void InitializePage ();
void ReadPage (UInt16 wPage);
void WritePage (UInt16 wPage);

void MakeRolls (FormPtr pForm);
void WriteRolls ();
void ClearRolls (Int16 iMaxRolls);
void InitRolls ();

Boolean CheckValue (UInt8 iCompareType, Int32 lCompareValue, Int32 lResult);
void PlayBeep (UInt8 iBeepType);

void ChangeCurrent (UInt16 wId);
void CustomizeButton (UInt16 wId);
void CheckCompare (DieRollPtr pDieRoll, Int32 lResult);

Boolean Initialize ();
Boolean Deinitialize ();

void AddPage ();
void DeletePage ();
void ClearPage ();

void GetButton (UInt16 wId, DieRollPtr drPtr);
void SetButton (UInt16 wId, DieRollPtr drPtr);

void DrawScreen (FormPtr pForm, Int16 iType);
void DrawControl (FormPtr pForm, Int16 iDisplayItem, Char *pLabel, RectangleType *pCtrlRect, Boolean bSelected);
void DrawIntValue (FormPtr pForm, Char *pLabel, Int32 iValue, RectangleType *pCtrlRect, Boolean bSelected);

Boolean ScreenTap (FormPtr pForm, EventPtr pEvent);
void ClickItem (FormPtr pForm, Int16 iDisplayItem);

RecordInfo g_vDataRecordInfo;
Boolean g_bDoCustom;
Char g_szResultBuffer [MAX_TEMPSTRING_LEN];
Char g_szCurrentBuffer [MAX_TEMPSTRING_LEN];
DieRollPtr g_pPage;
PrefsPtr g_pPrefs;
ResultsPtr g_pResults;
DieRollPtr g_pDieRoll;
DieRollStat g_vTotalRollStat;

UInt32 PilotMain (UInt16 cmd, void* cmdPBP, UInt16 launchFlags)
{
    Int16 error;

    if (cmd == sysAppLaunchCmdNormalLaunch)
    {
        error = StartApplication();	// Application start code
        if (error) return error;

        EventLoop();	// Event loop

        StopApplication ();	// Application stop code
    }
    return 0;
}

Int16 StartApplication()
{
	Int16 iError;

	iError = Initialize();
	if (iError) 
		return iError;
	iError = OpenDatabases();
	if (iError) 
		return iError;

	FrmGotoForm(IDF_MAIN);

	return iError;
}

Int16 StopApplication ()
{
	Int16 iError = 0;

	FrmSaveAllForms();
	FrmCloseAllForms();

	// Write the prefs and results.
	g_vDataRecordInfo.vInput.pRecord = g_pPrefs;
	g_vDataRecordInfo.vInput.vLength = sizeof(Prefs);
	g_vDataRecordInfo.vId = RECORDID_PREFS;
	g_vDataRecordInfo.vSubId = RECORDID_NONE;
	OverwriteRecordInfo(&g_vDataRecordInfo);
	CloseRecordInfo(&g_vDataRecordInfo);

	g_vDataRecordInfo.vInput.pRecord = g_pResults;
	g_vDataRecordInfo.vInput.vLength = sizeof(Results);
	g_vDataRecordInfo.vId = RECORDID_RESULTS;
	g_vDataRecordInfo.vSubId = RECORDID_NONE;
	OverwriteRecordInfo(&g_vDataRecordInfo);
	CloseRecordInfo(&g_vDataRecordInfo);

	// Save the current page before we leave.
	WritePage(g_pResults->wCurrentPage);

	RollFuncDestructor();
	CloseDatabases();
	Deinitialize();

	return iError;
}

Boolean OpenDatabases()
{
	RecordInfo vRecordInfo;

	// Create the roll database if one doesn't exist.
	if (OpenDatabaseRecordInfo(&vRecordInfo, ROLLEM_ID_APP, ROLLEM_DB_ROLL) == false)
	{
		if (CreateDatabaseRecordInfo(&vRecordInfo, ROLLEM_ID_APP, ROLLEM_DB_ROLL) == false)
		{
			return 1;
		}
	}
	CloseDatabaseRecordInfo(&vRecordInfo);

	// Open the data database.
	if (OpenDatabaseRecordInfo(&g_vDataRecordInfo, ROLLEM_ID_APP, ROLLEM_DB_DATA))
	{
		UInt16 wVersion = 0;

		// Check the version number.
		g_vDataRecordInfo.vId = RECORDID_PREFS;
		g_vDataRecordInfo.vSubId = RECORDID_NONE;
		if (ReadRecordInfo(&g_vDataRecordInfo))
		{
			MemMove(&wVersion, g_vDataRecordInfo.vDatabase.pRecord, sizeof(UInt16));
			CloseRecordInfo(&g_vDataRecordInfo);
		}

		if (wVersion < VERSION_NUMBER)
		{
			if (wVersion > 0)
			{
				if (FrmCustomAlert(IDA_CLEAR_DATABASE, NULL, NULL, NULL) == 1)
				{
					FrmCustomAlert(IDA_UPDATE_DATABASE, NULL, NULL, NULL);
					return 1;
				}
			}

			// Delete the old database.
			CloseDatabaseRecordInfo(&g_vDataRecordInfo);
			DeleteDatabaseRecordInfo(ROLLEM_ID_APP, ROLLEM_DB_DATA);

			return CreateDatabases();
		}

		// If we don't have the sounds record load a default.
		g_vDataRecordInfo.vId = RECORDID_SOUNDS;
		g_vDataRecordInfo.vSubId = RECORDID_NONE;
		if (ReadRecordInfo(&g_vDataRecordInfo))
		{
			CloseRecordInfo(&g_vDataRecordInfo);
		}
		else
		{
			LoadDefaultSounds();
		}

		// If we don't have the sounds record load a default.
		g_vDataRecordInfo.vId = RECORDID_DISPLAYS;
		g_vDataRecordInfo.vSubId = 0;
		if (ReadRecordInfo(&g_vDataRecordInfo))
		{
			CloseRecordInfo(&g_vDataRecordInfo);
		}
		else
		{
			LoadDefaultDisplays();
		}

		return 0;
	}

	return CreateDatabases();
}

Boolean CreateDatabases()
{
	if (CreateDatabaseRecordInfo(&g_vDataRecordInfo, ROLLEM_ID_APP, ROLLEM_DB_DATA))
		return InitDatabases();

	return 1;
}

Boolean CloseDatabases()
{
	if (g_pPrefs->wDeleteOnExit == 1)
		DeleteDatabaseRecordInfo(ROLLEM_ID_APP, ROLLEM_DB_ROLL);

	// Close the data database.
    CloseDatabaseRecordInfo(&g_vDataRecordInfo);

    return 0;
}

Boolean InitDatabases()
{
	MemHandle hSounds;

	g_vDataRecordInfo.vId = RECORDID_PREFS;
	g_vDataRecordInfo.vSubId = RECORDID_NONE;
	g_vDataRecordInfo.vInput.pRecord = g_pPrefs;
	g_vDataRecordInfo.vInput.vLength = sizeof(Prefs);
	InsertRecordInfo(&g_vDataRecordInfo);		// Prefs
	CloseRecordInfo(&g_vDataRecordInfo);

	g_vDataRecordInfo.vId = RECORDID_RESULTS;
	g_vDataRecordInfo.vSubId = RECORDID_NONE;
	g_vDataRecordInfo.vInput.pRecord = g_pResults;
	g_vDataRecordInfo.vInput.vLength = sizeof(Results);
	InsertRecordInfo(&g_vDataRecordInfo);		// Results
	CloseRecordInfo(&g_vDataRecordInfo);

	LoadDefaultSounds();
	LoadDefaultDisplays();
    InitializePage();

	g_vDataRecordInfo.vId = RECORDID_PAGES;
	for (g_vDataRecordInfo.vSubId=0; g_vDataRecordInfo.vSubId<g_pPrefs->wTotalPages; g_vDataRecordInfo.vSubId++)
	{
		g_vDataRecordInfo.vInput.pRecord = g_pPage;
		g_vDataRecordInfo.vInput.vLength = sizeof(DieRoll) * NUM_BUTTONS;
		InsertRecordInfo(&g_vDataRecordInfo);	// Pages
		CloseRecordInfo(&g_vDataRecordInfo);
	}

    return 0;
}

void EventLoop()
{
    short err;
    FormPtr pForm;
    EventType event;

    do
    {
        EvtGetEvent(&event, -1);

        if (SysHandleEvent(&event)) continue;
        if (MenuHandleEvent((void *)0, &event, &err)) continue;

        if (event.eType == frmLoadEvent)
        {
            pForm = FrmInitForm(event.data.frmLoad.formID);
            FrmSetActiveForm(pForm);
            switch (event.data.frmLoad.formID) 
            {
                case IDF_MAIN:
                    FrmSetEventHandler(pForm, (FormEventHandlerPtr) MainProc);
                    break;
                case IDF_PREFS:
                    FrmSetEventHandler(pForm, (FormEventHandlerPtr) PrefsProc);
                    break;
				case IDF_SOUND:
					FrmSetEventHandler(pForm, (FormEventHandlerPtr) SoundProc);
					break;
				case IDF_DISPLAY:
					FrmSetEventHandler(pForm, (FormEventHandlerPtr) DisplayProc);
					break;
            }
        }
        FrmDispatchEvent(&event);
    } while(event.eType != appStopEvent);
}

Boolean MainProc(EventPtr pEvent)
{
	FormPtr pForm = FrmGetActiveForm();
	Boolean bHandled = false;
#ifdef _DEBUG
    RectangleType rtRect;
	Char szBuffer [32];

    rtRect.topLeft.x = 55;
    rtRect.topLeft.y = 141;
    rtRect.extent.x = 20;
    rtRect.extent.y = 13;

    WinEraseRectangle(&rtRect, 0);
	StrLToA(szBuffer, (Int32)pEvent->eType);
    DrawCharacters(szBuffer, StrLen(szBuffer), &rtRect, DC_CENTER, true); 
#endif

    switch (pEvent->eType) 
    {
    case frmOpenEvent:
    {
		RollFuncConstructor();

		// Read the prefs and results.
		g_vDataRecordInfo.vId = RECORDID_PREFS;
		g_vDataRecordInfo.vSubId = RECORDID_NONE;
		if (ReadRecordInfo(&g_vDataRecordInfo))
		{
			MemMove(g_pPrefs, g_vDataRecordInfo.vDatabase.pRecord, g_vDataRecordInfo.vDatabase.vLength);
			CloseRecordInfo(&g_vDataRecordInfo);
		}

		g_vDataRecordInfo.vId = RECORDID_RESULTS;
		g_vDataRecordInfo.vSubId = RECORDID_NONE;
		if (ReadRecordInfo(&g_vDataRecordInfo))
		{
			MemMove(g_pResults, g_vDataRecordInfo.vDatabase.pRecord, g_vDataRecordInfo.vDatabase.vLength);
			CloseRecordInfo(&g_vDataRecordInfo);
		}

		// Right now we don't want to use Advance Custom dialog.
		g_pPrefs->wAdvCustom = 1;

		ReadPage(g_pResults->wCurrentPage);
		ChangeCurrent(g_pResults->wCurrentButton);

		RefreshForm();
		bHandled = true;
		break;
    }

	case frmUpdateEvent:
	{
        FrmDrawForm(pForm);
		DrawScreen(pForm, DRAW_ALL);
		bHandled = true;
		break;
	}

    case menuEvent:
	{
        switch (pEvent->data.menu.itemID)
        {
		case IDM_OPTIONS_ABOUT:
			ShowAbout();
			break;
		case IDM_OPTIONS_PREFS:
			FrmPopupForm(IDF_PREFS);
			break;
		case IDM_TOOLS_DISPLAY:
			FrmPopupForm(IDF_DISPLAY);
			break;
		case IDM_TOOLS_SOUND:
			FrmPopupForm(IDF_SOUND);
			break;
		case IDM_RECORD_ADD:
			AddPage();
			break;
		case IDM_RECORD_DELETE:
			DeletePage();
			break;
		case IDM_RECORD_CLEAR:
			ClearPage();
			break;
        }
        bHandled = true;
        break;
	}

	case keyDownEvent:
	{
        switch (pEvent->data.keyDown.chr)
        {
            case pageUpChr:
                PrevPage();
                break;
            case pageDownChr:
                NextPage();
                break;
        }
        break;
	}

	case penDownEvent:
	{
		bHandled = ScreenTap(pForm, pEvent);
		break;
	}

	default:
		bHandled = false;
    }
    return bHandled;
}

void CheckCompare (DieRollPtr pDieRoll, Int32 lResult)
{
    if (pDieRoll->lField[COMMON_BEEPTYPE1])
    {
        if (CheckValue(pDieRoll->lField[COMMON_BEEPCOMPARE1], pDieRoll->lField[COMMON_BEEPVALUE1], lResult))
        {
			PlayBeep(pDieRoll->lField[COMMON_BEEPTYPE1]);
        }
    }
    if (pDieRoll->lField[COMMON_BEEPTYPE2])
    {
        if (CheckValue(pDieRoll->lField[COMMON_BEEPCOMPARE2], pDieRoll->lField[COMMON_BEEPVALUE2], lResult))
        {
			PlayBeep(pDieRoll->lField[COMMON_BEEPTYPE2]);
        }
    }
}

Boolean CheckValue (UInt8 iCompareType, Int32 lCompareValue, Int32 lResult)
{
    switch(iCompareType)
    {
        case ComparisonEqual:
            return (lResult == lCompareValue);
        case ComparisonGreater:
            return (lResult > lCompareValue);
        case ComparisonLess:
            return (lResult < lCompareValue);
        case ComparisonGreaterEqual:
            return (lResult >= lCompareValue);
        case ComparisonLessEqual:
            return (lResult <= lCompareValue);
        case ComparisonNotEqual:
            return (lResult != lCompareValue);
        default:
            return false;
    }

}

void PlayBeep (UInt8 iBeepType)
{
	MemHandle hRecord;
	Char *pChar;
	SndCommandType sctSound;
	Int16 iCount;
 
	// Just making sure.
	if (iBeepType < 1)
		return;

	g_vDataRecordInfo.vId = RECORDID_SOUNDS;
	g_vDataRecordInfo.vSubId = RECORDID_NONE;
	if (ReadRecordInfo(&g_vDataRecordInfo))
	{
		pChar = (Char*)g_vDataRecordInfo.vDatabase.pRecord;
		iCount = *((UInt16*)pChar);

		iBeepType--; // Position 0 is No beep
		if ((iCount > 0) && (iBeepType < iCount))
		{
			// Moveit 
			pChar += sizeof(UInt16);
			while (iBeepType--)
			{
				// Skip the sounds.
				while (((NotePtr)pChar)->uiDuration)
					pChar += sizeof(Note);
				pChar += sizeof(Note);

				// Skip the titles.
				pChar += MAX_TITLE_LEN + 1;
			}

			// Play the sound.
			while (((NotePtr)pChar)->uiDuration)
			{
				PlayNote(((NotePtr)pChar)->uiHertz, ((NotePtr)pChar)->uiDuration);
				pChar += sizeof(Note);
			}
		}
		CloseRecordInfo(&g_vDataRecordInfo);
	}
}

void NextPage ()
{
    WritePage(g_pResults->wCurrentPage);
    g_pResults->wCurrentPage = (g_pResults->wCurrentPage + 1) % g_pPrefs->wTotalPages;
    ReadPage(g_pResults->wCurrentPage);

	// Refresh the page.
    DrawScreen(FrmGetActiveForm(), DRAW_PAGE);
}

void PrevPage ()
{
    WritePage(g_pResults->wCurrentPage);
    g_pResults->wCurrentPage = (!g_pResults->wCurrentPage) ? g_pPrefs->wTotalPages - 1 : g_pResults->wCurrentPage - 1; 
    ReadPage(g_pResults->wCurrentPage);
    DrawScreen(FrmGetActiveForm(), DRAW_PAGE);
}

void InitializePage ()
{
    Int16 iDefaultButtons [NUM_BUTTONS] = { 2, 4, 6, 8, 10, 12, 20, 100 };
    Int16 i;
    DieRollPtr pDieRoll;

	pDieRoll = g_pPage;
	MemSet(pDieRoll, (sizeof(DieRoll) * NUM_BUTTONS), '\0');
	for(i=IDC_BUTTON_FIRST; i<=IDC_BUTTON_LAST; i++)
	{
		pDieRoll->lField[COMMON_ROLLTYPE] = RollTypeNormal;
		pDieRoll->lField[NORMAL_DIE] = iDefaultButtons [ID2NUM(i)];
		pDieRoll->lField[NORMAL_NUMBER] = 1;

		CreateTitle(pDieRoll, pDieRoll->szTitle, MAX_TITLE_LEN+1);

		pDieRoll++;
	}
}

void ReadPage (UInt16 wPage)
{
	g_vDataRecordInfo.vId = RECORDID_PAGES;
	g_vDataRecordInfo.vSubId = wPage;
	if (ReadRecordInfo(&g_vDataRecordInfo))
	{
		MemMove(g_pPage, g_vDataRecordInfo.vDatabase.pRecord, g_vDataRecordInfo.vDatabase.vLength);
		CloseRecordInfo(&g_vDataRecordInfo);
	}
}

void WritePage (UInt16 wPage)
{
	g_vDataRecordInfo.vId = RECORDID_PAGES;
	g_vDataRecordInfo.vSubId = wPage;
	g_vDataRecordInfo.vInput.pRecord = g_pPage;
	g_vDataRecordInfo.vInput.vLength = sizeof(DieRoll) * NUM_BUTTONS;
	if (OverwriteRecordInfo(&g_vDataRecordInfo))
	{
		CloseRecordInfo(&g_vDataRecordInfo);
	}
}

void ChangeCurrent (UInt16 wId)
{
	if ((wId < IDC_BUTTON_FIRST) || 
		(wId > IDC_BUTTON_LAST))
		wId = IDC_BUTTON_FIRST;

	g_pResults->wCurrentButton = wId;
	GetButton(wId, g_pDieRoll);

	CreateTitle(g_pDieRoll, g_szCurrentBuffer, MAX_TEMPSTRING_LEN);
}

void GetButton (UInt16 wId, DieRollPtr drPtr)
{
	MemMove(drPtr, &g_pPage [ID2NUM(wId)], sizeof(DieRoll));
}

void SetButton (UInt16 wId, DieRollPtr drPtr)
{
	MemMove(&g_pPage [ID2NUM(wId)], drPtr, sizeof(DieRoll));
}

void CustomizeButton (UInt16 wId)
{
    if (wId == DisplayItemButtonCustom)
    {
        g_bDoCustom = !g_bDoCustom;            
    }
    else if (g_bDoCustom)
    {
		DieRoll vDieRoll;

		GetButton(wId, &vDieRoll);
		DoModalAdvCustom(&vDieRoll);
		SetButton(wId, &vDieRoll);
		g_bDoCustom = false;
    }

	DrawScreen(FrmGetActiveForm(), DRAW_PAGE);
}

Boolean Initialize ()
{
	g_pPrefs = MemPtrNew(sizeof(Prefs));
	g_pResults = MemPtrNew(sizeof(Results));
	g_pDieRoll = MemPtrNew(sizeof(DieRoll));
	g_pPage = MemPtrNew(sizeof(DieRoll) * NUM_BUTTONS);

	MemSet(g_pPrefs, sizeof(Prefs), '\0');
	MemSet(g_pResults, sizeof(Results), '\0');
	MemSet(g_pDieRoll, sizeof(DieRoll), '\0');

	// Initialize the prefs.
	g_pPrefs->wVersion = VERSION_NUMBER;
	g_pPrefs->wTotalPages = INIT_PAGES;
	g_pPrefs->wDeleteOnExit = INIT_ROLL_SESSION;
	g_pPrefs->wSaveRollNumber = INIT_ROLL_NUMBER;
	g_pPrefs->wAdvCustom = INIT_ADV_CUSTOM;
	g_pPrefs->wRollDisplayName = RollDisplayNameTitle;
	g_pPrefs->wRollDisplayStat = RollDisplayStatTotal;

	// Setup the database variables.
	InitializeRecordInfo(&g_vDataRecordInfo);

	// Initialize the roll structures.
	InitRolls();

	g_bDoCustom = false;

	return 0;
}

Boolean Deinitialize ()
{
	if (g_pPrefs)
		MemPtrFree(g_pPrefs);
	if (g_pResults)
		MemPtrFree(g_pResults);
	if (g_pDieRoll)
		MemPtrFree(g_pDieRoll);
	if (g_pPage)
		MemPtrFree(g_pPage);

	return 0;
}

void AddPage ()
{
	if (g_pPrefs->wTotalPages < MAX_PAGES)
	{
		g_pPrefs->wTotalPages++;

		// Save the current page before writeing a new one.
		WritePage(g_pResults->wCurrentPage);

		// Write and show the new page.
		InitializePage();

		g_pResults->wCurrentPage = g_pPrefs->wTotalPages - 1;
		WritePage(g_pResults->wCurrentPage);
		DrawScreen(FrmGetActiveForm(), DRAW_PAGE);
	}
}

void DeletePage ()
{
	UInt16 uiIndex;

	if (g_pPrefs->wTotalPages > MIN_PAGES)
	{
		if (!FrmCustomAlert(IDA_PAGE_CONFIRMATION, "delete", NULL, NULL))
		{
			// Delete the current page.
			g_vDataRecordInfo.vId = RECORDID_PAGES;
			g_vDataRecordInfo.vSubId = g_pResults->wCurrentPage;
			if (DeleteRecordInfo(&g_vDataRecordInfo) == true)
			{
				// Rename the pages after the one deleted.
				do 
				{
					g_vDataRecordInfo.vSubId++;
				} while (ChangeIdRecordInfo(&g_vDataRecordInfo, RECORDID_NONE, g_vDataRecordInfo.vSubId-1));

				// Reduce the number of pages.
				g_pPrefs->wTotalPages--;
				if (g_pResults->wCurrentPage >= g_pPrefs->wTotalPages)
					g_pResults->wCurrentPage = g_pPrefs->wTotalPages - 1;

				// Display the new page.
				ReadPage(g_pResults->wCurrentPage);
				DrawScreen(FrmGetActiveForm(), DRAW_PAGE);
			}
		}
	}
}

void ClearPage ()
{
    if (g_pPrefs->wTotalPages < MAX_PAGES)
    {
        if (!FrmCustomAlert(IDA_PAGE_CONFIRMATION, "clear", NULL, NULL))
        {
            InitializePage();
            WritePage(g_pResults->wCurrentPage);
            DrawScreen(FrmGetActiveForm(), DRAW_PAGE);
        }
    }
}

void MakeRolls (FormPtr pForm)
{
	// Roll the dice according to the type.
	RollDice(pForm, g_pDieRoll, g_pResults);

	// Do we beep?
	CheckCompare(g_pDieRoll, g_pResults->vLastRollStat.lTotal);
	WriteRolls();

	// Refresh the rolls.
	DrawScreen(FrmGetActiveForm(), DRAW_ROLL);
}

void WriteRolls ()
{
	RecordInfo vRecordInfo;

	ClearRolls(g_pPrefs->wSaveRollNumber - 1);

	if (OpenDatabaseRecordInfo(&vRecordInfo, ROLLEM_ID_APP, ROLLEM_DB_ROLL) == true)
	{
		UInt16 uiListSize = GetCurrentSizeList() * sizeof (Int16);

		// Use the index for saving rolls.
		vRecordInfo.vId = RECORDID_INDEX;
		vRecordInfo.vSubId = RECORDID_INDEX;

		// Insert a new record.
		vRecordInfo.vIndex = 0;
		vRecordInfo.vInput.pRecord = 0;
		vRecordInfo.vInput.vLength = sizeof(DieRoll) + sizeof(DieRollStat) + uiListSize;
		InsertRecordInfo(&vRecordInfo); // Rolls

		// Write the button.
		vRecordInfo.vInput.vOffset = 0;
		vRecordInfo.vInput.pRecord = g_pDieRoll;
		vRecordInfo.vInput.vLength = sizeof(DieRoll);
		WriteRecordInfo(&vRecordInfo); // Rolls

		// Write the stats.
		vRecordInfo.vInput.vOffset += vRecordInfo.vInput.vLength;
		vRecordInfo.vInput.pRecord = &g_pResults->vLastRollStat;
		vRecordInfo.vInput.vLength = sizeof(DieRollStat);
		WriteRecordInfo(&vRecordInfo); // Rolls

		// Write the rolls.
		vRecordInfo.vInput.vOffset += vRecordInfo.vInput.vLength;
		vRecordInfo.vInput.pRecord = GetList();
		vRecordInfo.vInput.vLength = uiListSize;
		WriteRecordInfo(&vRecordInfo); // Rolls
		ReleaseList();

		CloseRecordInfo(&vRecordInfo);
		
		// Close the database after wa are through with it.
		CloseDatabaseRecordInfo(&vRecordInfo);

		// Add the stats to the total.
		g_vTotalRollStat.lSize += g_pResults->vLastRollStat.lSize;
		g_vTotalRollStat.lTotalSize += g_pResults->vLastRollStat.lTotalSize;
		g_vTotalRollStat.lTotal += g_pResults->vLastRollStat.lTotal;
		g_vTotalRollStat.lSum += g_pResults->vLastRollStat.lSum;
		g_vTotalRollStat.lMean += g_pResults->vLastRollStat.lMean;
		g_vTotalRollStat.lMedian += g_pResults->vLastRollStat.lMedian;
		g_vTotalRollStat.lMode += g_pResults->vLastRollStat.lMode;
	}
}

void ClearRolls (Int16 iMaxRolls)
{
	RecordInfo vRecordInfo;

	if (OpenDatabaseRecordInfo(&vRecordInfo, ROLLEM_ID_APP, ROLLEM_DB_ROLL) == true)
	{
		UInt16 uiListSize = GetCurrentSizeList() * sizeof (Int16);

		// Cap the min at zero.
		if (iMaxRolls < 0)
			iMaxRolls = 0;

		// Use the index for saving rolls.
		vRecordInfo.vId = RECORDID_INDEX;
		vRecordInfo.vSubId = RECORDID_INDEX;
		vRecordInfo.vIndex = iMaxRolls;

		while (DmNumRecords(vRecordInfo.hDatabase) > iMaxRolls)
		{
			if (ReadRecordInfo(&vRecordInfo))
			{
				DieRollStat *pDieRollStat = (DieRollStat*)(((Char*)vRecordInfo.vDatabase.pRecord) + sizeof(DieRoll));

				// Subtract the stats from the total.
				g_vTotalRollStat.lSize -= pDieRollStat->lSize;
				g_vTotalRollStat.lTotalSize -= pDieRollStat->lTotalSize;
				g_vTotalRollStat.lTotal -= pDieRollStat->lTotal;
				g_vTotalRollStat.lSum -= pDieRollStat->lSum;
				g_vTotalRollStat.lMean -= pDieRollStat->lMean;
				g_vTotalRollStat.lMedian -= pDieRollStat->lMedian;
				g_vTotalRollStat.lMode -= pDieRollStat->lMode;

				CloseRecordInfo(&vRecordInfo);
			}

			DeleteRecordInfo(&vRecordInfo);
		}

		// Close the database after wa are through with it.
		CloseDatabaseRecordInfo(&vRecordInfo);
	}
}

void InitRolls ()
{
	RecordInfo vRecordInfo;
	Int16 iCount;

	// Initialize the variables.
	MemSet(&g_vTotalRollStat, sizeof(DieRollStat), '\0');

	if (OpenDatabaseRecordInfo(&vRecordInfo, ROLLEM_ID_APP, ROLLEM_DB_ROLL) == true)
	{
		iCount = DmNumRecords(vRecordInfo.hDatabase);

		// Use the index for saving rolls.
		vRecordInfo.vId = RECORDID_INDEX;
		vRecordInfo.vSubId = RECORDID_INDEX;

		for (vRecordInfo.vIndex=0; vRecordInfo.vIndex<iCount; vRecordInfo.vIndex++)
		{
			if (ReadRecordInfo(&vRecordInfo))
			{
				DieRollStat *pDieRollStat = (DieRollStat*)(((Char*)vRecordInfo.vDatabase.pRecord) + sizeof(DieRoll));

				// Subtract the stats from the total.
				g_vTotalRollStat.lSize += pDieRollStat->lSize;
				g_vTotalRollStat.lTotalSize += pDieRollStat->lTotalSize;
				g_vTotalRollStat.lTotal += pDieRollStat->lTotal;
				g_vTotalRollStat.lSum += pDieRollStat->lSum;
				g_vTotalRollStat.lMean += pDieRollStat->lMean;
				g_vTotalRollStat.lMedian += pDieRollStat->lMedian;
				g_vTotalRollStat.lMode += pDieRollStat->lMode;

				// Close the reocrd now that we are done with it.
				CloseRecordInfo(&vRecordInfo);
			}
		}

		// Close the database after wa are through with it.
		CloseDatabaseRecordInfo(&vRecordInfo);
	}
}

void DrawScreen (FormPtr pForm, Int16 iType)
{
	Boolean bVersion350 = IsRomVersion(VERSION_350);
	DisplayNode *pDisplayNode, *pRecordNode, vDisplayNode;
	Int16 iDisplayCount, i;
	Boolean bDraw;

	if (bVersion350 == true)
		WinPushDrawState();

	if (iType == DRAW_ALL)
	{
		RectangleType vCtrlRect;

		FrmGetObjectBounds(pForm, FrmGetObjectIndex(pForm, IDC_RESULT), &vCtrlRect);
		WinEraseRectangle(&vCtrlRect, 0);
	}

	g_vDataRecordInfo.vId = RECORDID_DISPLAYS;
	g_vDataRecordInfo.vSubId = 0;
	if (ReadRecordInfo(&g_vDataRecordInfo))
	{
		iDisplayCount = g_vDataRecordInfo.vDatabase.vLength / sizeof(DisplayNode);
		pRecordNode = (DisplayNode*)((Char*)g_vDataRecordInfo.vDatabase.pRecord + MAX_DISPLAY_TITLE);
		for (i=0; i<iDisplayCount; i++)
		{
			MemMove(&vDisplayNode, pRecordNode, sizeof(DisplayNode));

			vDisplayNode.vCtrlRect.topLeft.x += DISPLAY_OFFSET_X + 1;
			vDisplayNode.vCtrlRect.topLeft.y += DISPLAY_OFFSET_Y + 1;
			vDisplayNode.vCtrlRect.extent.x -= 2;
			vDisplayNode.vCtrlRect.extent.y -= 2;

			switch(vDisplayNode.vType)
			{
			case DisplayItemButtonRolls:
			case DisplayItemButtonClear:
			case DisplayItemButton0:
			case DisplayItemButton1:
			case DisplayItemButton2:
			case DisplayItemButton3:
			case DisplayItemButton4:
			case DisplayItemButton5:
			case DisplayItemButton6:
			case DisplayItemButton7:
			case DisplayItemButtonNext:
			case DisplayItemButtonPrev:
			case DisplayItemButtonPage:
			case DisplayItemButtonCustom:
				bDraw = ((iType & DRAW_PAGE) == DRAW_PAGE);
				break;
			case DisplayItemRollType:
			case DisplayItemRollTotal:
			case DisplayItemRollSum:
			case DisplayItemRollMean:
			case DisplayItemRollMedian:
			case DisplayItemRollMode:
			case DisplayItemRollSubDice:
			case DisplayItemRollSubRolls:
			case DisplayItemRollSubTotal:
				bDraw = ((iType & DRAW_ROLL) == DRAW_ROLL);
				break;
			default:
				bDraw = false;
			}

			if (bDraw == true)
			{
				DrawControl(pForm, vDisplayNode.vType, vDisplayNode.szTitle, &vDisplayNode.vCtrlRect, false);
			}

			pRecordNode++;
		}
		CloseRecordInfo(&g_vDataRecordInfo);
	}

	if (bVersion350 == true)
		WinPopDrawState();
	WinResetClip();
}

void DrawControl (FormPtr pForm, Int16 iDisplayItem, Char *pLabel, RectangleType *pCtrlRect, Boolean bSelected)
{
	Boolean bVersion350 = IsRomVersion(VERSION_350);
	Char szBuffer [32];
	FontID vFontId;
	DieRollPtr pDieRoll;

	switch(iDisplayItem)
	{
	case DisplayItemButtonRolls:
		DrawButtonValue(pForm, "R", pCtrlRect, bSelected);
		break;
	case DisplayItemButtonClear:
		DrawButtonValue(pForm, "C", pCtrlRect, bSelected);
		break;
	case DisplayItemButton0:
	case DisplayItemButton1:
	case DisplayItemButton2:
	case DisplayItemButton3:
	case DisplayItemButton4:
	case DisplayItemButton5:
	case DisplayItemButton6:
	case DisplayItemButton7:
		pDieRoll = g_pPage;
		pDieRoll += iDisplayItem - DisplayItemButton0;
		DrawButtonValue(pForm, pDieRoll->szTitle, pCtrlRect, bSelected);
		break;
	case DisplayItemButtonNext:
		vFontId = FntSetFont(3);
		DrawButtonValue(pForm, "\004", pCtrlRect, bSelected);
		FntSetFont(vFontId);
		break;
	case DisplayItemButtonPrev:
		vFontId = FntSetFont(3);
		DrawButtonValue(pForm, "\003", pCtrlRect, bSelected);
		FntSetFont(vFontId);
		break;
	case DisplayItemButtonPage:
		vFontId = FntSetFont(1);
		DrawIntValue(pForm, 0, (Int32)(g_pResults->wCurrentPage + 1), pCtrlRect, bSelected);
		FntSetFont(vFontId);
		break;
	case DisplayItemButtonCustom:
		vFontId = FntSetFont(1);
		DrawButtonValue(pForm, ((g_bDoCustom) ? "Cancel" : "Custom"), pCtrlRect, bSelected);
		FntSetFont(vFontId);
		break;
		break;
	case DisplayItemRollType:
		DrawStringValue(pForm, 0, g_szCurrentBuffer, pCtrlRect, bSelected);
		break;
	case DisplayItemRollTotal:
		DrawIntValue(pForm, pLabel, g_pResults->vLastRollStat.lTotal, pCtrlRect, bSelected);
		break;
	case DisplayItemRollSum:
		DrawIntValue(pForm, pLabel, g_pResults->vLastRollStat.lSum, pCtrlRect, bSelected);
		break;
	case DisplayItemRollMean:
		DrawIntValue(pForm, pLabel, g_pResults->vLastRollStat.lMean, pCtrlRect, bSelected);
		break;
	case DisplayItemRollMedian:
		DrawIntValue(pForm, pLabel, g_pResults->vLastRollStat.lMedian, pCtrlRect, bSelected);
		break;
	case DisplayItemRollMode:
		DrawIntValue(pForm, pLabel, g_pResults->vLastRollStat.lMode, pCtrlRect, bSelected);
		break;
	case DisplayItemRollSubDice:
		DrawIntValue(pForm, pLabel, g_vTotalRollStat.lSize, pCtrlRect, bSelected);
		break;
	case DisplayItemRollSubRolls:
	{
		RecordInfo vRecordInfo;
		Int16 iCount;

		if (OpenDatabaseRecordInfo(&vRecordInfo, ROLLEM_ID_APP, ROLLEM_DB_ROLL) == true)
		{
			iCount = DmNumRecords(vRecordInfo.hDatabase);
			// Close the database after we are through with it.
			CloseDatabaseRecordInfo(&vRecordInfo);
		}
		DrawIntValue(pForm, pLabel, iCount, pCtrlRect, bSelected);
		break;
	}
	case DisplayItemRollSubTotal:
		DrawIntValue(pForm, pLabel, g_vTotalRollStat.lTotal, pCtrlRect, bSelected);
		break;
	default:
		WinInvertRectangle(pCtrlRect, 0x0000);
	}
}

void DrawIntValue (FormPtr pForm, Char *pLabel, Int32 iValue, RectangleType *pCtrlRect, Boolean bSelected)
{
	Char szBuffer [32];

	StrLToA(szBuffer, iValue);
	DrawStringValue(pForm, pLabel, szBuffer, pCtrlRect, bSelected);
}

Boolean ScreenTap (FormPtr pForm, EventPtr pEvent)
{
	Boolean bVersion350 = IsRomVersion(VERSION_350);
	DisplayNode *pDisplayNode, *pRecordNode, vDisplayNode;
	Int16 iDisplayCount, i;
	Boolean bHandled = false;

	if (bVersion350 == true)
		WinPushDrawState();

	g_vDataRecordInfo.vId = RECORDID_DISPLAYS;
	g_vDataRecordInfo.vSubId = 0;
	if (ReadRecordInfo(&g_vDataRecordInfo))
	{
		iDisplayCount = g_vDataRecordInfo.vDatabase.vLength / sizeof(DisplayNode);
		pRecordNode = (DisplayNode*)((Char*)g_vDataRecordInfo.vDatabase.pRecord + MAX_DISPLAY_TITLE);
		for (i=0; i<iDisplayCount; i++)
		{
			MemMove(&vDisplayNode, pRecordNode, sizeof(DisplayNode));

			vDisplayNode.vCtrlRect.topLeft.x += DISPLAY_OFFSET_X + 1;
			vDisplayNode.vCtrlRect.topLeft.y += DISPLAY_OFFSET_Y + 1;
			vDisplayNode.vCtrlRect.extent.x -= 2;
			vDisplayNode.vCtrlRect.extent.y -= 2;

			// If we have 
			if (RctPtInRectangle(pEvent->screenX, pEvent->screenY, &vDisplayNode.vCtrlRect))
			{
				bHandled = true;
				break;
			}

			pRecordNode++;
		}
		CloseRecordInfo(&g_vDataRecordInfo);

		if (bHandled == true)
		{
			Int16 x, y;
			Boolean bPenDown;
			Boolean bSelected, bLastSelected;

			bLastSelected = false;
			do 
			{
				PenGetPoint(&x, &y, &bPenDown);
				bSelected = RctPtInRectangle(x, y, &vDisplayNode.vCtrlRect);
				if (bSelected != bLastSelected)
				{
					DrawControl(pForm, vDisplayNode.vType, vDisplayNode.szTitle, &vDisplayNode.vCtrlRect, bSelected);
					bLastSelected = bSelected;
				}
			} while (bPenDown);

			if (bSelected)
			{
				DrawControl(pForm, vDisplayNode.vType, vDisplayNode.szTitle, &vDisplayNode.vCtrlRect, false);
				ClickItem(pForm, vDisplayNode.vType); 
			}
		}
	}

	if (bVersion350 == true)
		WinPopDrawState();
	WinResetClip();

	return bHandled;
}

void ClickItem (FormPtr pForm, Int16 iDisplayItem)
{
	// Play the click sound.
	SndPlaySystemSound(sndClick);

	switch(iDisplayItem)
	{
	case DisplayItemButtonRolls:
		DoModalRollResults();
		break;
	case DisplayItemButtonClear:
		ClearRolls(0);
		DrawScreen(pForm, DRAW_ROLL);
		break;
	case DisplayItemButton0:
	case DisplayItemButton1:
	case DisplayItemButton2:
	case DisplayItemButton3:
	case DisplayItemButton4:
	case DisplayItemButton5:
	case DisplayItemButton6:
	case DisplayItemButton7:
		if (g_bDoCustom)
		{
			CustomizeButton(iDisplayItem);
		}
		else
		{
			ChangeCurrent(iDisplayItem);
			MakeRolls(pForm);
		}
		break;
	case DisplayItemButtonNext:
		NextPage();
		break;
	case DisplayItemButtonPrev:
		PrevPage();
		break;
	case DisplayItemButtonPage:
		break;
	case DisplayItemButtonCustom:
		CustomizeButton(iDisplayItem);
		break;
	case DisplayItemRollType:
	case DisplayItemRollTotal:
	case DisplayItemRollSum:
	case DisplayItemRollMean:
	case DisplayItemRollMedian:
	case DisplayItemRollMode:
	case DisplayItemRollSubDice:
	case DisplayItemRollSubRolls:
	case DisplayItemRollSubTotal:
		MakeRolls(pForm);
		break;
	default:;
	}
}

void ShowAbout ()
{
	Char szTitle [32];
	Char *pInfo;
	MemHandle hInfo;
	Boolean bStopBugging = false;

	// Get the strings from the resource.
	SysCopyStringResource(szTitle, IDS_ABOUT_TITLE);
	hInfo = DmGet1Resource('tSTR', IDS_ABOUT_INFO);

	DoInfoModal(szTitle, (Char*)MemHandleLock(hInfo), false, &bStopBugging);

	// Release the info string.
	MemHandleUnlock(hInfo);
	DmReleaseResource(hInfo);
}
