#include "StdInc.h"
#include "RollEm.h"

#include "Callback.h"
#include "EditNote.h"
#include "Sound.h"
#include "Strings.h"

Int16 g_iSoundPos, g_iNotePos;
Int16 g_iSoundSelect, g_iNoteSelect;
Int16 g_iSoundRows, g_iNoteRows;
StackNode g_vSounds;

void InitializeSound ();
void DeinitializeSound ();

void LoadSoundTable ();
void LoadNoteTable ();
void LoadRows ();

Int16 AddNote (UInt16 uiSoundIndex, UInt16 uiNoteIndex, NotePtr pNote);
Int16 GetNote (UInt16 uiSoundIndex, UInt16 uiNoteIndex, NotePtr *pNote);
Int16 RemoveNote (UInt16 uiSoundIndex, UInt16 uiNoteIndex);
Int16 GetNoteCount (UInt16 uiSoundIndex);

Int16 AddSound (UInt16 uiSoundIndex, SoundPtr pSound);
Int16 GetSound (UInt16 uiSoundIndex, SoundPtr *pSound);
Int16 RemoveSound (UInt16 uiSoundIndex);
Int16 GetSoundCount ();

void ReadSounds();
void WriteSounds();
void ClearSounds();

void SelectSound (Int16 iSound);
void SelectNote (Int16 iNote);

void MoveSoundList (Int16 iDirection);
void MoveNoteList (Int16 iDirection);

void RedrawSound ();
void RedrawNote ();

void TestSound ();

void PlayNote (UInt16 uiHertz, UInt16 uiDuration)
{
	SndCommandType sctSound;

	if ((uiHertz < 1) || (uiDuration < 1))
	{
		Int32 lTemp;

		lTemp = ((Int32)uiDuration * (Int32)SysTicksPerSecond()) / 1000;
		if (lTemp)
			SysTaskDelay(lTemp);
	}
	else
	{
		// Setup the default sound.
		sctSound.cmd = sndCmdFreqDurationAmp;
		sctSound.param3 = sndDefaultAmp;

		sctSound.param1 = uiHertz;
		sctSound.param2 = uiDuration;
		SndDoCmd((MemPtr)0, &sctSound, false);
	}
}

// Draw routine for the sounds.
void DrawSoundRecord (void *pTable, Int16 wRow, Int16 wColumn, RectangleType *pRect)
{
	UInt16 wLength;
	SoundPtr pSound;
	CALLBACK_PROLOGUE   /* Work around PilotGCC 0.5.0 bug */

	FntSetFont(stdFont);

	if (GetSound((wRow + g_iSoundPos), &pSound))
	{
		wLength = StrLen(pSound->szTitle);
		WinDrawChars(pSound->szTitle, wLength, pRect->topLeft.x + 5, pRect->topLeft.y);
	}

	CALLBACK_EPILOGUE /* Work around PilotGCC 0.5.0 bug */
}

// Draw routine for notes.
void DrawNoteRecord (void *pTable, Int16 wRow, Int16 wColumn, RectangleType *pRect)
{
	UInt16 wLength = 0;
	Char szBuffer [32];
	NotePtr pNote;
	CALLBACK_PROLOGUE   /* Work around PilotGCC 0.5.0 bug */

	FntSetFont(stdFont);

	if (GetNote(g_iSoundSelect, (wRow + g_iNotePos), &pNote))
	{
		StrLToA(szBuffer, pNote->uiHertz);
		wLength = StrLen(szBuffer);
		WinDrawChars(szBuffer, wLength, pRect->topLeft.x + 5, pRect->topLeft.y);

		StrLToA(szBuffer, pNote->uiDuration);
		wLength = StrLen(szBuffer);
		WinDrawChars(szBuffer, wLength, pRect->topLeft.x + 55, pRect->topLeft.y);
	}

	CALLBACK_EPILOGUE /* Work around PilotGCC 0.5.0 bug */
}

Boolean SoundProc(EventPtr pEvent)
{
	FormPtr pForm = FrmGetActiveForm();
	Boolean bHandled = false;

    switch (pEvent->eType) 
    {
    case frmOpenEvent:
    {
		InitializeSound();
		LoadSoundTable();
		RefreshForm();
		bHandled = true;
        break;
    }

    case frmCloseEvent:
    {
		DeinitializeSound();
		bHandled = false;
        break;
    }

	case frmUpdateEvent:
	{
		FrmDrawForm(pForm);
		RedrawSound();
		RedrawNote();
		bHandled = true;
		break;
	}

	case ctlSelectEvent:  // A control button was pressed and released.
	{
        switch (pEvent->data.ctlEnter.controlID)
        {
		case IDC_TEST_SOUND:
			TestSound();
			break;
		case IDC_INSERT_SOUND:
		{
			Char szTitle [MAX_TITLE_LEN];

			*szTitle = '\0';
			if (DoStringsModal("Title", szTitle, MAX_TITLE_LEN) == IDC_OK)
			{
				SoundPtr pSound;
				Int16 iCount;

				iCount = GetSoundCount();

				AddSound(iCount, 0);
				if (GetSound(iCount, &pSound))
				{
					g_iSoundSelect = iCount;
					g_iSoundPos = g_iSoundSelect - (g_iSoundRows - 1);
					if (g_iSoundPos < 0)
						g_iSoundPos = 0;

					StrCopy(pSound->szTitle, szTitle);
				}

				LoadRows();
				RedrawSound();
				RedrawNote();
			}
			break;
		}

		case IDC_DELETE_SOUND:
		{
		    if (!FrmCustomAlert(IDA_DELETE_CONFIRMATION, "sound", NULL, NULL))
		    {
				Int16 iCount;

				RemoveSound(g_iSoundSelect);

				iCount = GetSoundCount();
				if (g_iSoundSelect >= iCount)
					g_iSoundSelect = iCount - 1;
				g_iSoundPos = g_iSoundSelect - (g_iSoundRows - 1);
				if (g_iSoundPos < 0)
					g_iSoundPos = 0;

				LoadRows();
				RedrawSound();
				RedrawNote();
			}
			break;
		}

		case IDC_EDIT_SOUND:
		{
			SoundPtr pSound;

			if (GetSound(g_iSoundSelect, &pSound))
			{
				if (DoStringsModal("Title", pSound->szTitle, MAX_TITLE_LEN) == IDC_OK)
				{
					RedrawSound();
				}
			}
			break;
		}

		case IDC_INSERT_NOTE:
		{
			Note vNote;

			vNote.uiHertz = 0;
			vNote.uiDuration = 0;
			if (DoEditNote(&vNote) == IDC_OK)
			{
				Int16 iCount;

				iCount = GetNoteCount(g_iSoundSelect);

				g_iNoteSelect = iCount;
				g_iNotePos = g_iNoteSelect - (g_iNoteRows - 1);
				if (g_iNotePos < 0)
					g_iNotePos = 0;

				AddNote(g_iSoundSelect, iCount, &vNote);
				LoadRows();
				RedrawNote();
			}
			break;
		}

		case IDC_DELETE_NOTE:
		{
		    if (!FrmCustomAlert(IDA_DELETE_CONFIRMATION, "note", NULL, NULL))
		    {
				Int16 iCount;

				RemoveNote(g_iSoundSelect, g_iNoteSelect);
				
				iCount = GetNoteCount(g_iSoundSelect);
				if (g_iNoteSelect >= iCount)
					g_iNoteSelect = iCount - 1;
				g_iNotePos = g_iNoteSelect - (g_iNoteRows - 1);
				if (g_iNotePos < 0)
					g_iNotePos = 0;

				LoadRows();
				RedrawNote();
			}
			break;
		}

		case IDC_EDIT_NOTE:
		{
			NotePtr pNote;

			if (GetNote(g_iSoundSelect, g_iNoteSelect, &pNote))
			{
				DoEditNote(pNote);
				RedrawNote();
			}
			break;
		}
        case IDC_OK:
			WriteSounds();
        case IDC_CANCEL:
            FrmReturnToForm(0);
			DeinitializeSound();
            break;
        }
        bHandled = true;
        break;
	}

    case sclRepeatEvent:
	{
		switch(pEvent->data.sclRepeat.scrollBarID)
		{
		case IDC_NOTE_SCROLL:
			MoveNoteList(pEvent->data.sclRepeat.newValue - pEvent->data.sclRepeat.value);
			break;
		case IDC_SOUND_SCROLL:
			MoveSoundList(pEvent->data.sclRepeat.newValue - pEvent->data.sclRepeat.value);
			break;
		break;
		}
	}
	
	case tblSelectEvent:
	{
		switch(pEvent->data.tblEnter.tableID)
		{
		case IDC_SOUND_TABLE:
			SelectSound(pEvent->data.tblEnter.row);
			break;
		case IDC_NOTE_TABLE:
			SelectNote(pEvent->data.tblEnter.row);
			break;
		default:;
		}
		break;
	}

    case menuEvent:
	{
        switch (pEvent->data.menu.itemID)
        {
		case IDM_FORM_CLEAR:
		{
			if (!FrmCustomAlert(IDA_FORM_CLEAR, "sounds", NULL, NULL))
			{
				ClearSounds();
				LoadRows();
				RefreshForm();
			}
			break;
		}

		case IDM_FORM_DEFAULT:
		{
			if (!FrmCustomAlert(IDA_FORM_DEFAULT, "sounds", NULL, NULL))
			{
				ClearSounds();
				LoadDefaultSounds();
				ReadSounds();
				LoadRows();
				RefreshForm();
			}
			break;
		}

		default:;
        }
        bHandled = true;
        break;
	}

	default:
		bHandled = false;
    }

    return bHandled;
}

void LoadDefaultSounds ()
{
	MemHandle hDisplayItem;

	// Get the data from the resource.
	hDisplayItem = DmGet1Resource('Data', IDR_SOUNDS);
	if (hDisplayItem != 0)
	{
		g_vDataRecordInfo.vId = RECORDID_SOUNDS;
		g_vDataRecordInfo.vSubId = RECORDID_NONE;
		g_vDataRecordInfo.vInput.vOffset = 0;
		g_vDataRecordInfo.vInput.pRecord = MemHandleLock(hDisplayItem);
		g_vDataRecordInfo.vInput.vLength = MemHandleSize(hDisplayItem);

		if (OverwriteRecordInfo(&g_vDataRecordInfo))
		{
			CloseRecordInfo(&g_vDataRecordInfo);
		}

		// Release the data.
		MemHandleUnlock(hDisplayItem);
		DmReleaseResource(hDisplayItem);
	}
}

void InitializeSound ()
{
	// Do the every time initialization 
	//  of the Sound dialog here.
	StackConstruct(&g_vSounds);
	ReadSounds();

	g_iSoundPos = 0;
	g_iNotePos = 0;
	g_iSoundSelect = 0;
	g_iNoteSelect = 0;
}

void DeinitializeSound ()
{
	ClearSounds();
}

void LoadSoundTable ()
{
	FormPtr pForm = FrmGetActiveForm();
	TablePtr pControl;
	Int16 i, iRow;

	// Sound table.
	pControl = FrmGetObjectPtr(pForm, FrmGetObjectIndex(pForm, IDC_SOUND_TABLE));
	iRow = TblGetNumberOfRows(pControl);
	for (i=0; i<iRow; i++)
	{
		TblSetItemStyle(pControl, i, 0, customTableItem);
	}
	TblSetColumnUsable(pControl, 0, true);
	TblSetCustomDrawProcedure(pControl, 0, DrawSoundRecord);

	// Note table.
	pControl = FrmGetObjectPtr(pForm, FrmGetObjectIndex(pForm, IDC_NOTE_TABLE));
	iRow = TblGetNumberOfRows(pControl);
	for (i=0; i<iRow; i++)
	{
		TblSetItemStyle(pControl, i, 0, customTableItem);
	}
	TblSetColumnUsable(pControl, 0, true);
	TblSetCustomDrawProcedure(pControl, 0, DrawNoteRecord);

	LoadRows();
}

Int16 AddNote (UInt16 uiSoundIndex, UInt16 uiNoteIndex, NotePtr pNote)
{
	SoundPtr pSound;

	// Get the sound.
	if (GetSound(uiSoundIndex, &pSound) == 0)
		return 0;

	// Add a new note to the list.
	StackAddRear(&pSound->vNotes, pNote, sizeof(Note));

	return 1;
}

Int16 GetNote (UInt16 uiSoundIndex, UInt16 uiNoteIndex, NotePtr *pNote)
{
	SoundPtr pSound;

	// Get the sound.
	if (GetSound(uiSoundIndex, &pSound) == 0)
		return 0;

	// Do we have the record requested?
	if (StackGetCount(&pSound->vNotes) <= uiNoteIndex)
		return 0;

	*pNote = (NotePtr)StackGetData(&pSound->vNotes, uiNoteIndex);

	return 1;
}

Int16 RemoveNote (UInt16 uiSoundIndex, UInt16 uiNoteIndex)
{
	SoundPtr pSound;

	// Get the sound.
	if (GetSound(uiSoundIndex, &pSound) == 0)
		return 0;

	// Do we have the record requested?
	if (StackGetCount(&pSound->vNotes) <= uiNoteIndex)
		return 0;

	// Delete the requested record.
	StackDelete(&pSound->vNotes, uiNoteIndex);

	return 1;
}

Int16 GetNoteCount (UInt16 uiSoundIndex)
{
	SoundPtr pSound;

	// Get the sound.
	if (GetSound(uiSoundIndex, &pSound) == 0)
		return 0;

	return StackGetCount(&pSound->vNotes);
}

Int16 AddSound (UInt16 uiSoundIndex, SoundPtr pSound)
{
	SoundPtr pNewSound;
	Int16 iSoundCount;

	iSoundCount = StackGetCount(&g_vSounds);

	// Add a ne sound to the list.
	StackAddRear(&g_vSounds, pSound, sizeof(Sound));
	pNewSound = (SoundPtr)StackGetData(&g_vSounds, iSoundCount);

	// Clear out the stack.
	StackConstruct(&pNewSound->vNotes);

	return 1;
}

Int16 GetSound (UInt16 uiSoundIndex, SoundPtr *pSound)
{
	// Do we have the record requested?
	if (StackGetCount(&g_vSounds) <= uiSoundIndex)
		return 0;

	*pSound = (SoundPtr)StackGetData(&g_vSounds, uiSoundIndex);

	return 1;
}

Int16 RemoveSound (UInt16 uiSoundIndex)
{
	SoundPtr pSound;

	// Get the sound.
	if (GetSound(uiSoundIndex, &pSound) == 0)
		return 0;

	// Clear the notes.
	StackDestruct(&pSound->vNotes);

	// Delete the requested record.
	StackDelete(&g_vSounds, uiSoundIndex);

	return 1;
}

Int16 GetSoundCount ()
{
	return StackGetCount(&g_vSounds);
}

void ReadSounds ()
{
	Char *pChar;
	Int16 i, iCount, iNotes;
	SoundPtr pSound;

	iCount = 0;
	g_vDataRecordInfo.vId = RECORDID_SOUNDS;
	g_vDataRecordInfo.vSubId = RECORDID_NONE;
	if (ReadRecordInfo(&g_vDataRecordInfo))
	{
		pChar = (Char*)g_vDataRecordInfo.vDatabase.pRecord;
		iCount = *((UInt16*)pChar);

		if (iCount > 0)
		{
			// Moveit 
			pChar += sizeof(UInt16);
			for (i=0; i<iCount; i++)
			{
				AddSound(i, 0);
				GetSound(i, &pSound);

				// Add the notes.
				iNotes = 0;
				while (((NotePtr)pChar)->uiDuration)
				{
					AddNote(i, iNotes++, (NotePtr)pChar);
					pChar += sizeof(Note);
				}

				// Move up.
				pChar += sizeof(Note);

				// Copy the title.
				StrCopy(pSound->szTitle, pChar);

				// Skip the titles.
				pChar += MAX_TITLE_LEN + 1;
			}
		}

		CloseRecordInfo(&g_vDataRecordInfo);
	}
}

void WriteSounds()
{
	Char *pChar, *pTempChar;
    NotePtr pNote;
	SoundPtr pSound;
	Note vNote;
	Int16 i, iSize, iSounds, iNotes, iUntitled;

	if (g_vDataRecordInfo.hDatabase)
	{
		// Count the size needed.
		iUntitled = 1;
		iSounds = 0;
		iSize = sizeof(UInt16);

		while (GetSound(iSounds, &pSound) != 0)
		{
			Int16 iStrLen;

			iStrLen = StrLen(pSound->szTitle);
			while ((iStrLen > 0) && (pSound->szTitle [iStrLen-1] == ' '))
				iStrLen--;

			pSound->szTitle [iStrLen] = '\0';
			if (iStrLen == 0)
			{
				StrCopy(pSound->szTitle, "Untitled ");
				StrLToA((pSound->szTitle + 9), (Int32)iUntitled++);
			}

			iSize += MAX_TITLE_LEN + 1;
			iSize += (StackGetCount(&pSound->vNotes) + 1) * sizeof(Note);
			iSounds++;
		}

		// Initialize.
		MemSet(&vNote, sizeof(Note), 0);

		// Load the sounds into memory.
		pChar = (Char*)MemPtrNew(iSize);
		MemMove(pChar, &iSounds, sizeof(UInt16));
		pTempChar = pChar + sizeof(UInt16);
		for (i=0; i<iSounds; i++)
		{
			// Add the notes.
			iNotes = 0;
			while (GetNote(i, iNotes++, &pNote))
			{
				MemMove(pTempChar, pNote, sizeof(Note));
				pTempChar += sizeof(Note);
			}

			// Add the terminator..
			MemMove(pTempChar, &vNote, sizeof(Note));
			pTempChar += sizeof(Note);

			// Write the title.
			GetSound(i, &pSound);
			MemMove(pTempChar, pSound->szTitle, MAX_TITLE_LEN + 1);
			pTempChar += MAX_TITLE_LEN + 1;
		}

		// Write the new record.
		g_vDataRecordInfo.vId = RECORDID_SOUNDS;
		g_vDataRecordInfo.vSubId = RECORDID_NONE;
		g_vDataRecordInfo.vInput.pRecord = pChar;
		g_vDataRecordInfo.vInput.vLength = iSize;
		if (OverwriteRecordInfo(&g_vDataRecordInfo))
		{
			CloseRecordInfo(&g_vDataRecordInfo);
		}

		MemPtrFree(pChar);
	}
}

void ClearSounds ()
{
	while (RemoveSound(0));
}

void SelectSound (Int16 iSound)
{
	g_iSoundSelect = iSound + g_iSoundPos;
	g_iNotePos = 0;
	g_iNoteSelect = 0;
	LoadRows();
	RedrawNote();
}

void SelectNote (Int16 iNote)
{
	g_iNoteSelect = iNote + g_iNotePos;
}

void RedrawSound ()
{
	FormPtr pForm = FrmGetActiveForm();
	TablePtr pControl;
	Int16 iPos;

	pControl = FrmGetObjectPtr(pForm, FrmGetObjectIndex(pForm, IDC_SOUND_TABLE));
	TblEraseTable(pControl);
	TblDrawTable(pControl);

	if (GetSoundCount() > 0)
	{
		iPos = g_iSoundSelect - g_iSoundPos;
		if ((iPos >= 0) && (iPos < g_iSoundRows))
			TblSelectItem(pControl, iPos, 0);
	}

	MoveSoundList(0);
}

void RedrawNote ()
{
	FormPtr pForm = FrmGetActiveForm();
	TablePtr pControl;
	Int16 iPos;

	pControl = FrmGetObjectPtr(pForm, FrmGetObjectIndex(pForm, IDC_NOTE_TABLE));
	TblEraseTable(pControl);
	TblDrawTable(pControl);

	if (GetNoteCount(g_iSoundSelect) > 0)
	{
		iPos = g_iNoteSelect - g_iNotePos;
		if ((iPos >= 0) && (iPos < g_iNoteRows))
			TblSelectItem(pControl, iPos, 0);
	}

	MoveNoteList(0);
}

void TestSound ()
{
	Int16 i;
	NotePtr pNote;

	i = 0;
	while (GetNote(g_iSoundSelect, i++, &pNote))
		PlayNote(pNote->uiHertz, pNote->uiDuration);
}

void MoveSoundList (Int16 iDirection)
{
	FormPtr pForm = FrmGetActiveForm();
	ScrollBarPtr pScroll;
	UInt16 wUpIndex, wDownIndex;
	Boolean bScrollUp, bScrollDown;
	Int16 iNewPos, iCount;

	iCount = GetSoundCount();
	if (iDirection != 0)
	{
		iNewPos = g_iSoundPos + iDirection;

		if ((iNewPos + g_iSoundRows) > iCount)
			iNewPos = iCount - g_iSoundRows;
		if (iNewPos < 0)
			iNewPos = 0;

		if (iNewPos != g_iSoundPos)
		{
			g_iSoundPos = iNewPos;
			RedrawSound();
		}
	}

	iCount = (iCount > g_iSoundRows) ? iCount - g_iSoundRows : 0;

	pScroll = FrmGetObjectPtr(pForm, FrmGetObjectIndex(pForm, IDC_SOUND_SCROLL));
	SclSetScrollBar (pScroll, g_iSoundPos, 0, iCount, g_iSoundRows);
}

void MoveNoteList (Int16 iDirection)
{
	FormPtr pForm = FrmGetActiveForm();
	ScrollBarPtr pScroll;
	UInt16 wUpIndex, wDownIndex;
	Boolean bScrollUp, bScrollDown;
	Int16 iNewPos, i, iCount;

	iCount = GetNoteCount(g_iSoundSelect);
	if (iDirection != 0)
	{
		iNewPos = g_iNotePos + iDirection;

		if ((iNewPos + g_iNoteRows) > iCount)
			iNewPos = iCount - g_iNoteRows;
		if (iNewPos < 0)
			iNewPos = 0;

		if (iNewPos != g_iNotePos)
		{
			g_iNotePos = iNewPos;
			RedrawNote();
		}
	}

	iCount = (iCount > g_iSoundRows) ? iCount - g_iSoundRows : 0;

	pScroll = FrmGetObjectPtr(pForm, FrmGetObjectIndex(pForm, IDC_NOTE_SCROLL));
	SclSetScrollBar (pScroll, g_iNotePos, 0, iCount, g_iNoteRows);
}

void LoadRows ()
{
	FormPtr pForm = FrmGetActiveForm();
	TablePtr pControl;
	Int16 i, iCount;

	pControl = FrmGetObjectPtr(pForm, FrmGetObjectIndex(pForm, IDC_SOUND_TABLE));
	g_iSoundRows = TblGetNumberOfRows(pControl);
	iCount = GetSoundCount();
	for (i=0; i<g_iSoundRows; i++)
		TblSetRowUsable(pControl, i, (i < iCount) ? true : false);

	pControl = FrmGetObjectPtr(pForm, FrmGetObjectIndex(pForm, IDC_NOTE_TABLE));
	g_iNoteRows = TblGetNumberOfRows(pControl);
	iCount = GetNoteCount(g_iSoundSelect);
	for (i=0; i<g_iNoteRows; i++)
		TblSetRowUsable(pControl, i, (i < iCount) ? true : false);
}
