#include "StdInc.h"
#include "TabCtrl.h"

// Structs 
typedef struct _tagTabData
{
	UInt16 wItems;
	UInt16 wStart;
	UInt16 wSelected;
	Char *pItems;
	RectangleType vCtrlRect;
#ifdef _TABCALLBACK_
	TabCtrlFuncPtr pCallback;
#endif
} TabData;
typedef TabData *TabDataPtr;

// Defines
#define FRAME_TYPE	0x0301
#define TAB_TYPE	0x0201

#define TAB_LEFT	1
#define TAB_RIGHT	2
#define TAB_START	3

// Forward Declarations
Char *GetTabString (Char *pInput, UInt16 wIndex);
Int16 TabCtrlRectangle (TabDataPtr pTabData, Int16 iX, Int16 iY, RectanglePtr pCtrlRect, RectanglePtr pRect, Char **pString);
void TabCtrlDrawCtrl (FormPtr pForm, RectangleType *pCtrlRect, Char *pString, Int16 iType, Int16 iCurrent, Boolean bSelected);

void TabCtrlConstructor (FormPtr pForm, UInt16 wTabId, Char *pTabs)
{
	TabDataPtr pTabData;
	MemHandle hTabData;

	hTabData = MemHandleNew(sizeof(TabData));
	if (hTabData)
	{
		UInt16 wGadgetIndex = FrmGetObjectIndex(pForm, wTabId);
		RectangleType vCtrlRect;
		Char *pTemp;

		pTabData = (TabDataPtr)MemHandleLock(hTabData);
		pTabData->wItems = 0;
		pTabData->wStart = 0;
		pTabData->wSelected = 0;
		pTabData->pItems = (Char*)MemPtrNew(StrLen(pTabs) + 1);

#ifdef _TABCALLBACK_
		pTabData->pCallback = 0;
#endif
		StrCopy(pTabData->pItems, pTabs);
		pTemp = pTabData->pItems;
		while (*pTemp)
		{
			if (*pTemp == '|')
			{
				*pTemp = '\0';
				pTabData->wItems++;
			}
			pTemp++;
		}

		// Get the complete ctrol size.
		FrmGetObjectBounds(pForm, wGadgetIndex, &pTabData->vCtrlRect);

		// Create the tab rectangle.
		vCtrlRect.topLeft.x = pTabData->vCtrlRect.topLeft.x;
		vCtrlRect.topLeft.y = pTabData->vCtrlRect.topLeft.y; 
		vCtrlRect.extent.x = pTabData->vCtrlRect.extent.x;
		vCtrlRect.extent.y = FntLineHeight();
		FrmSetObjectBounds(pForm, wGadgetIndex, &vCtrlRect);

		MemHandleUnlock(hTabData);
		FrmSetGadgetData(pForm, FrmGetObjectIndex(pForm, wTabId), hTabData);
	}
}

void TabCtrlDestructor (FormPtr pForm, UInt16 wTabId)
{
	TabDataPtr pTabData;
	MemHandle hTabData;

	hTabData = FrmGetGadgetData(pForm, FrmGetObjectIndex(pForm, wTabId));
	pTabData = (TabDataPtr)MemHandleLock(hTabData);
	MemPtrFree(pTabData->pItems);
	MemHandleUnlock(hTabData);
	MemHandleFree(hTabData);
}

#ifdef _TABCALLBACK_
void TabCtrlSetCallback(FormPtr pForm, UInt16 wTabId, TabCtrlFuncPtr pCallback)
{
	TabDataPtr pTabData;
	MemHandle hTabData;

	hTabData = FrmGetGadgetData(pForm, FrmGetObjectIndex(pForm, wTabId));

	pTabData = (TabDataPtr)MemHandleLock(hTabData);
	pTabData->pCallback = pCallback;
	MemHandleUnlock(hTabData);
}
#endif

void TabCtrlSetUsable (FormPtr pForm, UInt16 wTabId, Boolean bUsable)
{
	GadgetPtr pGadget;

	pGadget = FrmGetObjectPtr(pForm, FrmGetObjectIndex(pForm, wTabId));
	pGadget->attr.usable = bUsable;
	TabCtrlDraw(pForm, wTabId);
}

Boolean TabCtrlGetUsable (FormPtr pForm, UInt16 wTabId)
{
	GadgetPtr pGadget;

	pGadget = FrmGetObjectPtr(pForm, FrmGetObjectIndex(pForm, wTabId));
	return pGadget->attr.usable;
}

void TabCtrlDrawCtrl (FormPtr pForm, RectangleType *pCtrlRect, Char *pString, Int16 iType, Int16 iCurrent, Boolean bSelected)
{
	Boolean bVersion350 = IsRomVersion(VERSION_350);
	IndexedColorType vForeColorType, vBackColorType;

	if (iType == TABCTRL_NOTAP)
		return;

	if (bVersion350 == true)
	{
		if ((iCurrent == iType) || (iType < TAB_START))
		{
			if (bSelected == true)
			{
				vBackColorType = UIColorGetTableEntryIndex(UIObjectSelectedFill);
				vForeColorType = UIColorGetTableEntryIndex(UIObjectSelectedForeground);
			}
			else
			{
				vBackColorType = UIColorGetTableEntryIndex(UIObjectFill);
				vForeColorType = UIColorGetTableEntryIndex(UIObjectForeground);
			}
		}
		else
		{
			if (bSelected == true)
			{
				vBackColorType = UIColorGetTableEntryIndex(UIObjectSelectedFill);
				vForeColorType = UIColorGetTableEntryIndex(UIObjectSelectedForeground);
			}
			else
			{
				RGBColorType vRGBColorType;

				vRGBColorType.r = 204;
				vRGBColorType.g = 204;
				vRGBColorType.b = 204;
				vBackColorType = WinRGBToIndex(&vRGBColorType);
				vRGBColorType.r = 0;
				vRGBColorType.g = 0;
				vRGBColorType.b = 0;
				vForeColorType = WinRGBToIndex(&vRGBColorType);
			}
		}
	}

	if (iType == TAB_LEFT)
	{
		Int16 iX, iY, i;

		iX = pCtrlRect->topLeft.x;
		iY = pCtrlRect->topLeft.y + (pCtrlRect->extent.y / 2);

		if (bVersion350 == true)
		{
			WinSetForeColor(vBackColorType);
			WinDrawRectangle(pCtrlRect, 0);
			WinSetForeColor(vForeColorType);
		}

		for (i=0; i<4; i++)
		{
			WinDrawLine(iX+i, iY-i, iX+i, iY+i);
		}
	}
	else if (iType == TAB_RIGHT)
	{
		Int16 iX, iY, i;

		iX = pCtrlRect->topLeft.x + pCtrlRect->extent.x - 1;
		iY = pCtrlRect->topLeft.y + (pCtrlRect->extent.y / 2);

		if (bVersion350 == true)
		{
			WinSetForeColor(vBackColorType);
			WinDrawRectangle(pCtrlRect, 0);
			WinSetForeColor(vForeColorType);
		}

		for (i=0; i<4; i++)
		{
			WinDrawLine(iX-i, iY-i, iX-i, iY+i);
		}
	}
	else
	{
		// Set the clip rect.
		InflateRect(pCtrlRect, 1, 1);
		WinSetClip(pCtrlRect);
		InflateRect(pCtrlRect, -1, -1);

		if (bVersion350 == true)
		{
			WinSetForeColor(UIColorGetTableEntryIndex(UIObjectFrame));
		}

		// Draw the border.
		WinDrawRectangleFrame(TAB_TYPE, pCtrlRect);

		if (bVersion350 == true)
		{
			WinSetBackColor(vBackColorType);
			WinSetForeColor(vBackColorType);
			WinDrawRectangle(pCtrlRect, 0);
		}

		if (iCurrent == iType)
		{
			WinEraseLine(
				pCtrlRect->topLeft.x,
				pCtrlRect->topLeft.y + pCtrlRect->extent.y,
				pCtrlRect->topLeft.x + pCtrlRect->extent.x - 1, 
				pCtrlRect->topLeft.y + pCtrlRect->extent.y);
		}

		// Draw the string.
		if (pString != 0)
		{
			if (bVersion350 == true)
			{
				WinSetTextColor(vForeColorType);
			}
			WinSetClip(pCtrlRect);
			WinDrawChars(pString, StrLen(pString), pCtrlRect->topLeft.x + 1, pCtrlRect->topLeft.y - 1);
		}
	}
}

void TabCtrlDraw(FormPtr pForm, UInt16 wTabId)
{
	Boolean bVersion350 = IsRomVersion(VERSION_350);
	UInt16 wGadgetIndex = FrmGetObjectIndex(pForm, wTabId);
	RectangleType vCtrlRect, vTempRect;
	GadgetPtr pGadget;
	TabDataPtr pTabData;
	MemHandle hTabData;

	hTabData = FrmGetGadgetData(pForm, FrmGetObjectIndex(pForm, wTabId));
	pTabData = (TabDataPtr)MemHandleLock(hTabData);

	FrmGetObjectBounds(pForm, wGadgetIndex, &vCtrlRect);
	MemMove(&vTempRect, &pTabData->vCtrlRect, sizeof(RectangleType));

	// Modify the bottom rectangle.
	vTempRect.topLeft.y += FntLineHeight() + 1;
	vTempRect.extent.y -= FntLineHeight() + 1;

	InflateRect(&vCtrlRect, 1, 1);
	WinEraseRectangle(&vCtrlRect, 0);
	InflateRect(&vCtrlRect, -1, -1);

	pGadget = FrmGetObjectPtr(pForm, wGadgetIndex);
	if (pGadget->attr.usable)
	{
		FontID vOldFont;
		Int16 i;

		if (bVersion350 == true)
		{
			// Save the draw state.
			WinPushDrawState();
		}
		WinDrawRectangleFrame(FRAME_TYPE, &vTempRect);

		vTempRect.topLeft.x = vCtrlRect.topLeft.x + 2;
		for (i=pTabData->wStart; i<pTabData->wItems; i++)
		{
			Char *pText;
			UInt16 wSize;

			pText = GetTabString(pTabData->pItems, i);

			wSize = FntCharsWidth(pText, StrLen(pText));
			if (pTabData->wSelected == i)
			{
				vTempRect.topLeft.y = vCtrlRect.topLeft.y;
				vTempRect.extent.y = vCtrlRect.extent.y;
			}
			else
			{
				vTempRect.topLeft.y = vCtrlRect.topLeft.y + 1;
				vTempRect.extent.y = vCtrlRect.extent.y - 1;
			}
			vTempRect.extent.x = wSize + 1;
			if ((vTempRect.topLeft.x + vTempRect.extent.x) >= (vCtrlRect.topLeft.x + vCtrlRect.extent.x - 10))
				break;

			TabCtrlDrawCtrl(pForm, &vTempRect, pText, i + TAB_START, pTabData->wSelected + TAB_START, false);
			vTempRect.topLeft.x += vTempRect.extent.x + 1;
		}
		WinResetClip();

		// Not tabs are shown.
		if (pTabData->wStart != 0)
		{
			Int16 iX, iY, i;
			Char cTemp;

			vTempRect.topLeft.x = vCtrlRect.topLeft.x + vCtrlRect.extent.x - 10;
			vTempRect.topLeft.y = vCtrlRect.topLeft.y;
			vTempRect.extent.x = 5;
			vTempRect.extent.y = 11;
			TabCtrlDrawCtrl(pForm, &vTempRect, 0, TAB_LEFT, 0, false);
		}

		if (i < pTabData->wItems)
		{
			Int16 iX, iY, i;
			Char cTemp;

			vTempRect.topLeft.x = vCtrlRect.topLeft.x + vCtrlRect.extent.x - 5;
			vTempRect.topLeft.y = vCtrlRect.topLeft.y;
			vTempRect.extent.x = 5;
			vTempRect.extent.y = 11;
			TabCtrlDrawCtrl(pForm, &vTempRect, 0, TAB_RIGHT, 0, false);
		}

		if (bVersion350 == true)
		{
			// Restore the draw state.
			WinPopDrawState();
		}
	}
	else
	{
		WinEraseRectangleFrame(FRAME_TYPE, &vTempRect);
	}

	MemHandleUnlock(hTabData);
}

Int16 TabCtrlRectangle (TabDataPtr pTabData, Int16 iX, Int16 iY, RectanglePtr pCtrlRect, RectanglePtr pRect, Char **pString)
{
	UInt16 wSize;
	Int16 i;

	pRect->topLeft.x = pCtrlRect->topLeft.x + 2;

	for (i=pTabData->wStart; i<pTabData->wItems; i++)
	{
		*pString = GetTabString(pTabData->pItems, i);
		wSize = FntCharsWidth(*pString, StrLen(*pString));
		if (pTabData->wSelected == i)
		{
			pRect->topLeft.y = pCtrlRect->topLeft.y;
			pRect->extent.y = pCtrlRect->extent.y;
		}
		else
		{
			pRect->topLeft.y = pCtrlRect->topLeft.y + 1;
			pRect->extent.y = pCtrlRect->extent.y - 1;
		}
		pRect->extent.x = wSize + 1;

		if ((pRect->topLeft.x + pRect->extent.x) >= (pCtrlRect->topLeft.x + pCtrlRect->extent.x - 10))
			break;
		if (RctPtInRectangle(iX, iY, pRect))
			return i + TAB_START;
		pRect->topLeft.x += pRect->extent.x + 1;
	}

	if (pTabData->wStart != 0)
	{
		pRect->topLeft.x = pCtrlRect->topLeft.x + pCtrlRect->extent.x - 10;
		pRect->topLeft.y = pCtrlRect->topLeft.y;
		pRect->extent.x = 5;
		pRect->extent.y = pCtrlRect->extent.y;
		if (RctPtInRectangle(iX, iY, pRect))
			return TAB_LEFT;
	}

	if (i < pTabData->wItems)
	{
		pRect->topLeft.x = pCtrlRect->topLeft.x + pCtrlRect->extent.x - 5;
		pRect->topLeft.y = pCtrlRect->topLeft.y;
		pRect->extent.x = 5;
		pRect->extent.y = pCtrlRect->extent.y;
		if (RctPtInRectangle(iX, iY, pRect))
			return TAB_RIGHT;
	}

	pRect->extent.x = 0;
	pRect->extent.y = 0;

	return TABCTRL_NOTAP;
}

Int16 TabCtrlTap(FormPtr pForm, UInt16 wTabId, EventPtr pEvent)
{
	Boolean bVersion350 = IsRomVersion(VERSION_350);
	UInt16 wGadgetIndex = FrmGetObjectIndex(pForm, wTabId);
	Int16 iReturn = TABCTRL_NOTAP;
	GadgetPtr pGadget;
	RectangleType vCtrlRect, vLastRect, vTempRect;
	TabDataPtr pTabData;
	MemHandle hTabData;
	Int16 iResult, iLastResult;
	Int16 x, y;
	Boolean bPenDown;
	Char *pString, *pLastString;
#ifdef _TABCALLBACK_
	Int16 iCurrent, iTotal;
	TabCtrlFuncPtr pCallback;
#endif

	pGadget = FrmGetObjectPtr(pForm, wGadgetIndex);
	if (!pGadget->attr.usable)
		return iReturn;

	FrmGetObjectBounds(pForm, wGadgetIndex, &vCtrlRect);
	if (!RctPtInRectangle(pEvent->screenX, pEvent->screenY, &vCtrlRect))
		return iReturn;

	hTabData = FrmGetGadgetData(pForm, FrmGetObjectIndex(pForm, wTabId));
	pTabData = (TabDataPtr)MemHandleLock(hTabData);

#ifdef _TABCALLBACK_
	// Get the current stats of the tabctrl.
	iCurrent = pTabData->wSelected;
	iTotal = pTabData->wItems;
	pCallback = pTabData->pCallback;
#endif

	vLastRect.extent.x = 0;
	vLastRect.extent.y = 0;
	iLastResult = TABCTRL_NOTAP;
	pLastString = 0;

	if (bVersion350 == true)
	{
		// Save the draw state.
		WinPushDrawState();
	}

	do 
	{
		PenGetPoint(&x, &y, &bPenDown);
		iResult = TabCtrlRectangle (pTabData, x, y, &vCtrlRect, &vTempRect, &pString);
		if (iLastResult != iResult)
		{
			if (bVersion350 == true)
			{
				TabCtrlDrawCtrl(pForm, &vLastRect, pLastString, iLastResult, pTabData->wSelected + TAB_START, false);
				TabCtrlDrawCtrl(pForm, &vTempRect, pString, iResult, pTabData->wSelected + TAB_START, true);
			}
			else
			{
				WinInvertRectangle(&vLastRect, 0);
				WinInvertRectangle(&vTempRect, 0);
			}
			vLastRect.topLeft.x = vTempRect.topLeft.x;
			vLastRect.topLeft.y = vTempRect.topLeft.y;
			vLastRect.extent.x = vTempRect.extent.x;
			vLastRect.extent.y = vTempRect.extent.y;
			iLastResult = iResult;
			pLastString = pString;
		}
	} while (bPenDown);

	if (bVersion350 == true)
	{
		TabCtrlDrawCtrl(pForm, &vLastRect, pLastString, iLastResult, pTabData->wSelected + TAB_START, false);
		WinPopDrawState();
	}
	else
	{
		WinInvertRectangle(&vLastRect, 0);
	}
	WinResetClip();

	switch (iResult)
	{
	case TABCTRL_NOTAP:
		break;
	case TAB_LEFT:
		pTabData->wStart--;
		TabCtrlDraw(pForm, wTabId);
		SndPlaySystemSound(sndClick);
		break;
	case TAB_RIGHT:
		pTabData->wStart++;
		TabCtrlDraw(pForm, wTabId);
		SndPlaySystemSound(sndClick);
		break;
	default:
		if (pTabData->wSelected != iResult - TAB_START)
		{
			pTabData->wSelected = iResult - TAB_START;
			TabCtrlDraw(pForm, wTabId);
			iReturn = pTabData->wSelected;
		}
		SndPlaySystemSound(sndClick);
	}

	MemHandleUnlock(hTabData);

#ifdef _TABCALLBACK_
	// If we have a callback function call it.
	if (pCallback != 0)
	{
		pCallBack(pForm, wTabId, iTotal, iCurrent, iReturn);
	}
#endif

	return iReturn;
}

Int16 TabCtrlTotal (FormPtr pForm, UInt16 wTabId)
{
	TabDataPtr pTabData;
	MemHandle hTabData;
	Int16 iCount;

	hTabData = FrmGetGadgetData(pForm, FrmGetObjectIndex(pForm, wTabId));
	pTabData = (TabDataPtr)MemHandleLock(hTabData);
	iCount = pTabData->wItems;
	MemHandleUnlock(hTabData);

	return iCount;
}

Int16 TabCtrlCurrent (FormPtr pForm, UInt16 wTabId)
{
	TabDataPtr pTabData;
	MemHandle hTabData;
	Int16 iCount;

	hTabData = FrmGetGadgetData(pForm, FrmGetObjectIndex(pForm, wTabId));
	pTabData = (TabDataPtr)MemHandleLock(hTabData);
	iCount = pTabData->wSelected;
	MemHandleUnlock(hTabData);

	return iCount;
}

Int16 TabCtrlStart (FormPtr pForm, UInt16 wTabId)
{
	TabDataPtr pTabData;
	MemHandle hTabData;
	Int16 iCount;

	hTabData = FrmGetGadgetData(pForm, FrmGetObjectIndex(pForm, wTabId));
	pTabData = (TabDataPtr)MemHandleLock(hTabData);
	iCount = pTabData->wSelected;
	MemHandleUnlock(hTabData);

	return iCount;
}

void TabCtrlSetCurrent (FormPtr pForm, UInt16 wTabId, Int16 iCurrent)
{
	TabDataPtr pTabData;
	MemHandle hTabData;

	hTabData = FrmGetGadgetData(pForm, FrmGetObjectIndex(pForm, wTabId));
	pTabData = (TabDataPtr)MemHandleLock(hTabData);
	pTabData->wSelected = iCurrent;
	MemHandleUnlock(hTabData);
}

void TabCtrlSetStart (FormPtr pForm, UInt16 wTabId, Int16 iStart)
{
	TabDataPtr pTabData;
	MemHandle hTabData;

	hTabData = FrmGetGadgetData(pForm, FrmGetObjectIndex(pForm, wTabId));
	pTabData = (TabDataPtr)MemHandleLock(hTabData);
	pTabData->wStart = iStart;
	MemHandleUnlock(hTabData);
}

Char *GetTabString (Char *pInput, UInt16 wIndex)
{
	Char *pTemp = pInput;

	while (wIndex-- && *pTemp)
	{
		pTemp += StrLen(pTemp) + 1;
	}

	return pTemp;
}
