#include "StdInc.h"
#include "RollEmUpd.h"

Int16 StartApplication();
Int16 StopApplication();
void EventLoop();

Boolean Initialize ();
Boolean Deinitialize ();

void InitializePage (DieRoll101Ptr pDieRoll101);
void ReadPage (UInt16 wPage, DieRoll101Ptr pDieRoll101);
void WritePage (UInt16 wPage, DieRoll101Ptr pDieRoll101);

void CreateTitle (DieRoll101Ptr pDieRoll101, Char *charPtr, UInt16 wLen);
Boolean CheckValue (UInt8 iCompareType, Int32 lCompareValue, Int32 lResult);
void PlayBeep (UInt8 iBeepType);

void ChangeCurrent (UInt16 wId);
void CustomizeButton (UInt16 wId);
void CheckCompare (DieRoll101 dr, Int32 lResult);

void AddPage ();
void DeletePage ();
void ClearPage ();

Boolean FindVersion ();
Boolean ConvertDatabase ();
Boolean ConvertDatabase000to110 ();
Boolean ConvertDatabase101to110 ();
Boolean ConvertDatabase110to200 ();

Boolean OverwriteIndex (RecordInfoPtr pRecordInfo);

void ShowNumber (Char *pTitle, Int32 lNumber);
void ShowString (Char *pTitle, Char *pString);

// ************************************************************* //

UInt16 g_wVersion;
RecordInfo g_vRecordInfo;

// ************************************************************* //

UInt32 PilotMain (UInt16 cmd, void* cmdPBP, UInt16 launchFlags)
{
    Int16 error;

    if (cmd == sysAppLaunchCmdNormalLaunch)
    {
        error = StartApplication();	// Application start code
        if (error) return error;
//		EventLoop();	// Event loop
        StopApplication ();	// Application stop code
    }

    return 0;
}

Int16 StartApplication()
{
	Int16 error;

	if (error = Initialize())
		return error;

	ConvertDatabase();

	return error;
}

Int16 StopApplication ()
{
	Int16 iError = 0;

	FrmSaveAllForms();
	FrmCloseAllForms();

    Deinitialize();

	return iError;
}

void EventLoop()
{
    short err;
    Int16 formID;
    FormPtr fpForm;
    EventType event;

    do
    {
        EvtGetEvent(&event, -1);

        if (SysHandleEvent(&event)) continue;
        if (MenuHandleEvent((void *)0, &event, &err)) continue;

        if (event.eType == frmLoadEvent)
        {
            formID = event.data.frmLoad.formID;
            fpForm = FrmInitForm(formID);
            FrmSetActiveForm(fpForm);
            switch (formID) 
            {
			default:;
            }
        }
        FrmDispatchEvent(&event);
    } while(event.eType != appStopEvent);
}

Boolean Initialize ()
{
    return 0;
}

Boolean Deinitialize ()
{
    return 0;
}

Boolean FindVersion ()
{
	Int16 iCount, i;
	Boolean bDatabaseInfo = false;

	g_wVersion = 0;

	// Try to open it or create it.
	if (OpenDatabaseRecordInfo(&g_vRecordInfo, ROLLEM_ID_APP, ROLLEM_DB_DATA) == false)
	{
		if (CreateDatabaseRecordInfo(&g_vRecordInfo, ROLLEM_ID_APP, ROLLEM_DB_DATA) == false)
		{
			return false;
		}
	}

	// Get the total records.
	iCount = DmNumRecords(g_vRecordInfo.hDatabase);
	for (i=0; i<iCount; i++)
	{
		MemHandle hRecord;

		hRecord = DmGetRecord(g_vRecordInfo.hDatabase, i);
		if (hRecord != 0)
		{
			if (MemHandleSize(hRecord) >= 8)
			{
				Int32 *pLong;

				pLong = (Int32*)MemHandleLock(hRecord);
				if ((*(pLong+0) == RECORDID_PREFS) &&
					(*(pLong+1) == RECORDID_NONE))
				{
					bDatabaseInfo = true;
				}
				MemHandleUnlock(hRecord); 
			}
		}
		DmReleaseRecord(g_vRecordInfo.hDatabase, i, false);
	}

	ShowNumber("DatabaseInfo", (Int32)bDatabaseInfo);
	if (bDatabaseInfo == true)
	{
		g_vRecordInfo.vId = RECORDID_PREFS;
		g_vRecordInfo.vSubId = RECORDID_NONE;
		if (ReadRecordInfo(&g_vRecordInfo))
		{
			MemMove(&g_wVersion, g_vRecordInfo.vDatabase.pRecord, sizeof(UInt16));
			CloseRecordInfo(&g_vRecordInfo);
		}
	}
	else
	{
		ReadRecord(g_vRecordInfo.hDatabase, PREFS_RECORD, &g_wVersion, sizeof(UInt16));
	}

	CloseDatabaseRecordInfo(&g_vRecordInfo);

	return true;
}

Boolean ConvertDatabase ()
{
	if (FindVersion() == true)
	{
		ShowNumber("Version", g_wVersion);
		if (g_wVersion < 101)
		{
			ConvertDatabase000to110();
		}

		ShowNumber("Version", g_wVersion);
		if (g_wVersion < 110)
		{
			ConvertDatabase101to110();
		}
		
		ShowNumber("Version", g_wVersion);
		if (g_wVersion < 200)
		{
			ConvertDatabase110to200();
		}
	}

	ShowNumber("Version", g_wVersion);
	FrmCustomAlert(IDA_CONVERT_FINISH, NULL, NULL, NULL);

	return true;;
}

Boolean ConvertDatabase000to110 ()
{
    UInt16 uiIndex;
    UInt32 ulLength;
    Int16 i;
	DieRoll101Ptr pDieRoll;
	Prefs101 vPrefs101;

	OpenDatabaseRecordInfo(&g_vRecordInfo, ROLLEM_ID_APP, ROLLEM_DB_DATA);

	// This is the base database make a clean state.
	if (g_vRecordInfo.hDatabase)
	{
		UInt16 uiRecords;

		// Delete all records.
		uiRecords = DmNumRecords(g_vRecordInfo.hDatabase);
		while (uiRecords--)
			DmRemoveRecord(g_vRecordInfo.hDatabase, 0);
	}

	// Create the prefs record.
	// Initialize the prefs
	MemSet(&vPrefs101, sizeof(Prefs101), '\0');
	vPrefs101.wVersion = 101;
	vPrefs101.wTotalPages = DEFAULT_PAGES;

	// Write the prefs out to the database
	InsertRecord(g_vRecordInfo.hDatabase, PREFS_RECORD, &vPrefs101, sizeof(Prefs101));

	// Create the page records
    pDieRoll = MemPtrNew(sizeof(DieRoll101) * NUM_BUTTONS);
	if (pDieRoll)
	{
		// Initialize the page
	    InitializePage(pDieRoll);
		for (i=0; i<DEFAULT_PAGES; i++)
		{
			// Write the prefs out to the database
			InsertRecord(g_vRecordInfo.hDatabase, PAGE2REC101(i), pDieRoll, (sizeof(DieRoll101) * NUM_BUTTONS));
		}
	}

	CloseDatabaseRecordInfo(&g_vRecordInfo);

	// We now have a 101 database set the version version number
	g_wVersion = 101;

    return 0;
}

void InitializePage (DieRoll101Ptr pDieRoll101)
{
    Int16 iDefaultButtons [NUM_BUTTONS] = { 2, 4, 6, 8, 10, 12, 20, 100 };
    Int16 i;

    MemSet(pDieRoll101, (sizeof(DieRoll101) * NUM_BUTTONS), '\0');
    for(i=IDC_BUTTON_FIRST; i<=IDC_BUTTON_LAST; i++)
    {
        pDieRoll101->iDie = iDefaultButtons [ID2NUM(i)];
        pDieRoll101->iNumber = 1;
        pDieRoll101->iModifier = 0;

        CreateTitle(pDieRoll101, pDieRoll101->szTitle, MAX_TITLE_LEN);

        pDieRoll101++;
    }
}

void CreateTitle (DieRoll101Ptr pDieRoll101, Char *charPtr, UInt16 wLen)
{
	Char szTemp[MAX_TEMPSTRING_LEN];

	*szTemp = '\0';
    if ((pDieRoll101->iBest) && (pDieRoll101->iBest < pDieRoll101->iNumber))
    {
        StrCat(szTemp, "Best ");
        StrIToA(&szTemp[StrLen(szTemp)], pDieRoll101->iBest);
        StrCat(szTemp, " of ");
        StrIToA(&szTemp[StrLen(szTemp)], pDieRoll101->iNumber);
    }
    else
    {
        StrIToA(&szTemp[StrLen(szTemp)], pDieRoll101->iNumber);
    }
    StrCat(szTemp, "d");
    StrIToA(&szTemp[StrLen(szTemp)], pDieRoll101->iDie);
    if ((pDieRoll101->iModifier > 0) && (pDieRoll101->lSuccessValue == 0))
    {
        if (pDieRoll101->iModifier > 0)
            StrCat(szTemp, "+");
        StrIToA(&szTemp[StrLen(szTemp)], pDieRoll101->iModifier);
    }
/*
	if (pDieRoll101->lSuccessValue > 0)
	{
		StrCat(szTemp, " ");
		StrCat(szTemp, g_szCompare[pDieRoll101->iSuccessCompare]);
		StrCat(szTemp, " ");
		StrIToA(&szTemp[StrLen(szTemp)], pDieRoll101->lSuccessValue);
	}
*/
	if (wLen > MAX_TEMPSTRING_LEN)
		wLen = MAX_TEMPSTRING_LEN;
	StrNCopy(charPtr, szTemp, --wLen);
	charPtr[wLen] = '\0';
}

Boolean ConvertDatabase101to110 ()
{
	UInt16 wTemp;
	UInt16 uiIndex;
	Prefs101 vPrefs101;
	Prefs110 vPrefs110;
	Results110 vResults110;
	DieRoll101Ptr pDieRoll101;
	DieRoll110Ptr pDieRoll110;
	Int16 i, t;

	OpenDatabaseRecordInfo(&g_vRecordInfo, ROLLEM_ID_APP, ROLLEM_DB_DATA);

	ReadRecord(g_vRecordInfo.hDatabase, PREFS_RECORD, &vPrefs101, sizeof(Prefs101));

	vPrefs110.wVersion = 110;
    vPrefs110.wTotalPages = vPrefs101.wTotalPages;

	OverwriteRecord(g_vRecordInfo.hDatabase, PREFS_RECORD, &vPrefs110, sizeof(Prefs110));

    vResults110.wCurrentPage = vPrefs101.wCurrentPage;
    vResults110.wCurrentButton = vPrefs101.wCurrentButton;
	vResults110.lLastResult = 0;
	vResults110.lLastRolls [1] = 0;

	InsertRecord(g_vRecordInfo.hDatabase, RESULTS_RECORD, &vResults110, sizeof(Results110));

	wTemp = 0;
	uiIndex = SOUNDS_RECORD;
	InsertRecord(g_vRecordInfo.hDatabase, uiIndex++, &wTemp, sizeof(UInt16));
	InsertRecord(g_vRecordInfo.hDatabase, uiIndex++, &wTemp, sizeof(UInt16));
	InsertRecord(g_vRecordInfo.hDatabase, uiIndex++, &wTemp, sizeof(UInt16));
	InsertRecord(g_vRecordInfo.hDatabase, uiIndex++, &wTemp, sizeof(UInt16));
	InsertRecord(g_vRecordInfo.hDatabase, uiIndex++, &wTemp, sizeof(UInt16));
	InsertRecord(g_vRecordInfo.hDatabase, uiIndex++, &wTemp, sizeof(UInt16));
	InsertRecord(g_vRecordInfo.hDatabase, uiIndex++, &wTemp, sizeof(UInt16));
	InsertRecord(g_vRecordInfo.hDatabase, uiIndex++, &wTemp, sizeof(UInt16));

	pDieRoll101 = MemPtrNew(sizeof(DieRoll101) * NUM_BUTTONS);
	pDieRoll110 = MemPtrNew(sizeof(DieRoll110) * NUM_BUTTONS);

	if (pDieRoll101 && pDieRoll110)
	{
		for (i=0; i<vPrefs110.wTotalPages; i++)
		{
			ReadRecord(g_vRecordInfo.hDatabase, PAGE2REC110(i), pDieRoll101, (sizeof(DieRoll101) * NUM_BUTTONS));

			// Clear the record
			MemSet(pDieRoll110, (sizeof(DieRoll110) * NUM_BUTTONS), '\0');

			for (t=0; t<NUM_BUTTONS; t++)
			{
				// Copy the record.
				StrCopy(pDieRoll110 [t].szTitle, pDieRoll101 [t].szTitle);
				pDieRoll110 [t].lField00 = pDieRoll101 [t].iDie;
				pDieRoll110 [t].lField01 = pDieRoll101 [t].iNumber;
				pDieRoll110 [t].lField02 = pDieRoll101 [t].iModifier;
				pDieRoll110 [t].lField03 = pDieRoll101 [t].iBest;
				pDieRoll110 [t].lField04 = pDieRoll101 [t].iBeepType1;
				pDieRoll110 [t].lField05 = pDieRoll101 [t].iBeepType2;
				pDieRoll110 [t].lField06 = pDieRoll101 [t].iBeepCompare1;
				pDieRoll110 [t].lField07 = pDieRoll101 [t].iBeepCompare2;
				pDieRoll110 [t].lField08 = pDieRoll101 [t].iSuccessCompare;
				pDieRoll110 [t].lField09 = pDieRoll101 [t].lBeepValue1;
				pDieRoll110 [t].lField10 = pDieRoll101 [t].lBeepValue2;
				pDieRoll110 [t].lField11 = pDieRoll101 [t].lSuccessValue;
			}

			OverwriteRecord(g_vRecordInfo.hDatabase, PAGE2REC110(i), pDieRoll110, (sizeof(DieRoll110) * NUM_BUTTONS));
		}
	}

    if (pDieRoll101)
        MemPtrFree(pDieRoll101);
    if (pDieRoll110)
        MemPtrFree(pDieRoll110);

	CloseDatabaseRecordInfo(&g_vRecordInfo);

	// We now have a 110 database set the version version number
	g_wVersion = 110;
}

Boolean ConvertDatabase110to200 ()
{
	Prefs110 vPrefs110;
	Prefs200 vPrefs200;
	Results110 vResults110;
	Results200 vResults200;
	MemHandle hRecord;
	Char *pRecord;
	Int16 iCount, i;

	MemSet(&vPrefs200, sizeof(Prefs200), '\0');
	MemSet(&vResults200, sizeof(Results200), '\0');

	OpenDatabaseRecordInfo(&g_vRecordInfo, ROLLEM_ID_APP, ROLLEM_DB_DATA);

// ************************************************************* //

	// Read and convert in the old records before database conversion.
	ReadRecord(g_vRecordInfo.hDatabase, PREFS_RECORD, &vPrefs110, sizeof(Prefs110));

	vPrefs200.wVersion = 200;
    vPrefs200.wTotalPages = vPrefs110.wTotalPages;
	vPrefs200.wDeleteOnExit = INIT_ROLL_SESSION;
	vPrefs200.wSaveRollNumber = INIT_ROLL_NUMBER;
	vPrefs200.wAdvCustom = INIT_ADV_CUSTOM;
	vPrefs200.wRollDisplayName = RollDisplayNameTitle;
	vPrefs200.wRollDisplayStat = RollDisplayStatTotal;

	ReadRecord(g_vRecordInfo.hDatabase, RESULTS_RECORD, &vResults110, sizeof(Results110));

    vResults200.wCurrentPage = vResults110.wCurrentPage;
    vResults200.wCurrentButton = vResults110.wCurrentButton;
	vResults200.lLastResult = vResults110.lLastResult;

// ************************************************************* //

	iCount = DmNumRecords(g_vRecordInfo.hDatabase);
	for (i=0; i<iCount; i++)
	{
		g_vRecordInfo.vIndex = i;

		if (i == PREFS_RECORD)
		{
			g_vRecordInfo.vId = RECORDID_PREFS;
			g_vRecordInfo.vSubId = RECORDID_NONE;
		}
		else if (i == RESULTS_RECORD)
		{
			g_vRecordInfo.vId = RECORDID_RESULTS;
			g_vRecordInfo.vSubId = RECORDID_NONE;
		}
		else if (i == SOUNDS_RECORD)
		{
			g_vRecordInfo.vId = RECORDID_SOUNDS;
			g_vRecordInfo.vSubId = RECORDID_NONE;
		}
		else if ((i > SOUNDS_RECORD) && (i < PAGES_RECORD))
		{
			g_vRecordInfo.vId = RECORDID_NONE;
			g_vRecordInfo.vSubId = RECORDID_NONE;
		}
		else
		{
			g_vRecordInfo.vId = RECORDID_PAGES;
			g_vRecordInfo.vSubId = i - PAGES_RECORD;
		}

		OverwriteIndex(&g_vRecordInfo);
	}

// ************************************************************* //

	g_vRecordInfo.vId = RECORDID_PREFS;
	g_vRecordInfo.vSubId = RECORDID_NONE;
	g_vRecordInfo.vInput.pRecord = &vPrefs200;
	g_vRecordInfo.vInput.vLength = sizeof(Prefs200);
	OverwriteRecordInfo(&g_vRecordInfo);
	CloseRecordInfo(&g_vRecordInfo);

	// Write out the new records.
	g_vRecordInfo.vId = RECORDID_RESULTS;
	g_vRecordInfo.vSubId = RECORDID_NONE;
	g_vRecordInfo.vInput.pRecord = &vResults200;
	g_vRecordInfo.vInput.vLength = sizeof(Results200);
	OverwriteRecordInfo(&g_vRecordInfo);
	CloseRecordInfo(&g_vRecordInfo);

// ************************************************************* //

	DmQuickSort(g_vRecordInfo.hDatabase, RecordInfoSort, 0);

	g_vRecordInfo.vId = RECORDID_NONE;
	g_vRecordInfo.vSubId = RECORDID_NONE;
	DeleteRecordInfo (&g_vRecordInfo);

	CloseDatabaseRecordInfo(&g_vRecordInfo);

	// We now have a 110 database set the version version number
	g_wVersion = 200;
}

Boolean OverwriteIndex (RecordInfoPtr pRecordInfo)
{
	MemHandle hRecord;

	hRecord = DmGetRecord(pRecordInfo->hDatabase, pRecordInfo->vIndex);
	pRecordInfo->vInput.pRecord = MemHandleLock(hRecord);
	pRecordInfo->vInput.vLength = MemHandleSize(hRecord);

	// Write the new record after the old record.
	pRecordInfo->vIndex++;
	pRecordInfo->vDatabase.hRecord = DmNewRecord(pRecordInfo->hDatabase, &pRecordInfo->vIndex, (pRecordInfo->vInput.vLength + 8));
	pRecordInfo->vDatabase.pRecord = MemHandleLock(pRecordInfo->vDatabase.hRecord);
	pRecordInfo->vDatabase.vLength = pRecordInfo->vInput.vLength;

	// Write the record id
	DmWrite(pRecordInfo->vDatabase.pRecord, 0, &pRecordInfo->vId, 4);
	// Write the record sub id
	DmWrite(pRecordInfo->vDatabase.pRecord, 4, &pRecordInfo->vSubId, 4);
	// Write the record data
	if (pRecordInfo->vInput.pRecord != 0)
		DmWrite(pRecordInfo->vDatabase.pRecord, 8, pRecordInfo->vInput.pRecord, pRecordInfo->vInput.vLength);

	MemHandleUnlock(pRecordInfo->vDatabase.hRecord); 
	DmReleaseRecord(pRecordInfo->hDatabase, pRecordInfo->vIndex, false);
	pRecordInfo->vIndex--;

	// Get rid of the old record.
	MemHandleUnlock(hRecord); 
	DmReleaseRecord(pRecordInfo->hDatabase, pRecordInfo->vIndex, false);
	DmRemoveRecord(pRecordInfo->hDatabase, pRecordInfo->vIndex);

	return true;
}

void ShowNumber (Char *pTitle, Int32 lNumber)
{
	Char szBuffer [32];

	StrLToA(szBuffer, lNumber);
	ShowString(pTitle, szBuffer);
}

void ShowString (Char *pTitle, Char *pString)
{
#ifdef _DEBUG
	FrmCustomAlert(IDA_DEBUG, pTitle, pString, NULL);
#endif
}
